from __future__ import annotations

import json
import os
from getpass import getpass
from pathlib import Path
from typing import Dict, NamedTuple, Optional, Union

from elabapi_python import Configuration  # type: ignore
from tabulate import tabulate

from elabftwcontrol.defaults import DEFAULT_CONFIG_FILE

Pathlike = Union[str, Path]


class AccessConfig(NamedTuple):
    host_url: str
    api_key: str
    verify_ssl: bool
    debug: bool = False

    def get_api_config(self) -> Configuration:
        configuration = Configuration()
        configuration.api_key["api_key"] = self.api_key
        configuration.api_key_prefix["api_key"] = "Authorization"
        configuration.host = self.host_url
        configuration.debug = self.debug
        configuration.verify_ssl = self.verify_ssl
        return configuration


class MultiConfig(NamedTuple):
    data: Dict[str, AccessConfig]

    def has_profile(self, profile: str) -> bool:
        return profile in self.data

    def get_profile(self, profile: str) -> AccessConfig:
        return self.data[profile]

    def set_profile(self, profile: str, accessconfig: AccessConfig) -> None:
        self.data[profile] = accessconfig

    def delete_profile(self, profile: str) -> None:
        del self.data[profile]

    @classmethod
    def from_file(cls, filepath: Pathlike) -> MultiConfig:
        """Read data from a json file"""
        data: Dict[str, AccessConfig] = {}
        with open(filepath, "r") as f:
            for profile, access_parameters in json.load(f).items():
                data[profile] = AccessConfig(**access_parameters)
        return cls(data)

    def to_file(self, filepath: Pathlike) -> None:
        """Export configuration to a json file"""
        with open(filepath, "w") as f:
            f.write(
                json.dumps(
                    {
                        profile: accessconfig._asdict()
                        for profile, accessconfig in self.data.items()
                    }
                )
            )


def yes_no(
    prompt: str,
    default_yes: bool = True,
) -> bool:
    if default_yes:
        return not input(f"{prompt} [Y|n]").lower().strip().startswith("n")
    else:
        return input(f"{prompt} [y|N]").lower().strip().startswith("y")


def create_or_append_configuration_file(
    filepath: Pathlike,
    profile: Optional[str] = None,
    show_keys: bool = False,
) -> None:
    if profile is None:
        profile = "default"
    pass_prompt = "Enter API key: "
    if show_keys:
        api_key = input(pass_prompt).strip()
    else:
        api_key = getpass(pass_prompt).strip()

    host = input("Enter elabAPI endpoint: ").strip()

    verify_ssl = yes_no("Verify SSL?", default_yes=True)

    debug = yes_no("Debug mode?", default_yes=False)

    new_config = AccessConfig(
        host_url=host,
        api_key=api_key,
        verify_ssl=verify_ssl,
        debug=debug,
    )

    if Path(filepath).exists():
        config = MultiConfig.from_file(filepath)
    else:
        config = MultiConfig({})

    if config.has_profile(profile):
        if not yes_no(f"Overwrite existing profile '{profile}'?", default_yes=False):
            print("Aborted.")
            return

    config.set_profile(profile, new_config)
    config.to_file(filepath)

    print(f"Wrote configuration to {filepath}.")


def delete_configuration_file(
    filepath: Pathlike,
    profile: Optional[str] = None,
    show_keys: bool = False,
) -> None:
    if Path(filepath).exists():
        config = MultiConfig.from_file(filepath)
    else:
        print("Config file does not exist.")
        return

    if profile is None:
        if not yes_no(f"Delete config file '{filepath}'?", default_yes=False):
            print("Aborted.")
            return
        os.remove(filepath)
        print(f"Deleted configuration file {filepath}")
    else:
        if not config.has_profile(profile):
            print(f"Profile {profile} does not exist.")
        if not yes_no(f"Delete config profile '{profile}'?", default_yes=False):
            print("Aborted.")
            return
        config.delete_profile(profile)
        config.to_file(filepath)
        print(f"Deleted configuration profile {profile}")


def list_config_profiles(
    filepath: Pathlike,
    profile: Optional[str] = None,
    show_keys: bool = False,
) -> None:
    if not Path(filepath).exists():
        print(f"Config file '{filepath}' does not exist. Can not list profiles.")
        return

    config = MultiConfig.from_file(filepath)
    table = []
    for read_profile, single_config in config.data.items():
        if profile is None or profile == read_profile:
            key = single_config.api_key if show_keys else "***"
            row = [
                read_profile,
                single_config.host_url,
                key,
                single_config.verify_ssl,
                single_config.debug,
            ]
            table.append(row)
    print(
        tabulate(
            table,
            headers=["Profile", "Endpoint", "API key", "Verify SSL", "Debug mode"],
        )
    )


def main() -> None:
    create_or_append_configuration_file(DEFAULT_CONFIG_FILE)


if __name__ == "__main__":
    main()



from __future__ import annotations

from typing import Iterable, Iterator, List, Literal, NamedTuple, Optional

import pandas as pd

try:
    import awswrangler as wr
    from awswrangler.typing import _S3WriteDataReturnValue
    from s3fs import S3FileSystem  # type: ignore

    S3_ENABLED = True
except ImportError:
    S3_ENABLED = False


from elabftwcontrol._logging import logger


class ParsedS3Path(NamedTuple):
    bucket: str
    prefix: str

    @classmethod
    def from_path(cls, path: str) -> ParsedS3Path:
        if not path.startswith("s3://"):
            raise ValueError("Path must start with s3://")

        stripped_path = path[5:]
        split_path = stripped_path.split("/")
        bucket = split_path[0]
        prefix = "/".join(split_path[1:])
        return cls(
            bucket=bucket,
            prefix=prefix,
        )

    @property
    def path(self) -> str:
        return f"s3://{self.bucket}/{self.prefix}"

    @property
    def stripped_path(self) -> str:
        return f"{self.bucket}/{self.prefix}"


def _assert_s3_enabled() -> None:
    assert S3_ENABLED, "You have to install the optional S3 dependencies"


def write_df_to_glue_table(
    df: pd.DataFrame,
    path: str,
    database: str,
    table: str,
    partitions: Optional[List[str]] = None,
) -> _S3WriteDataReturnValue:
    _assert_s3_enabled()
    mode: Literal["overwrite", "overwrite_partitions"]
    if partitions is not None:
        mode = "overwrite_partitions"
    else:
        mode = "overwrite"

    return wr.s3.to_parquet(
        df=df,
        path=path,
        index=False,
        sanitize_columns=True,
        dataset=True,
        partition_cols=partitions,
        mode=mode,
        database=database,
        table=table,
    )


def write_df_to_s3_parquet(
    bucket: str,
    prefix: str,
    dataframe: pd.DataFrame,
) -> None:
    """Write a dataframe without registering it in Glue"""
    _assert_s3_enabled()
    wr.s3.to_parquet(
        df=dataframe,
        path=ParsedS3Path(bucket=bucket, prefix=prefix).path,
        compression="snappy",
    )


def write_lines_s3(
    bucket: str,
    prefix: str,
    lines: Iterable[str],
) -> None:
    _assert_s3_enabled()
    s3 = S3FileSystem()
    # hack to count items, because python doesn't have pointers to ints
    number_of_lines = [0]

    def process_substring(substr: str) -> bytes:
        number_of_lines[0] += 1
        return (substr + "\n").encode("UTF-8")

    with s3.open(f"{bucket}/{prefix}", mode="wb") as f:
        f.writelines(
            (process_substring(line) for line in lines),
        )

    logger.info(
        "Wrote %s lines to bucket %s on prefix %s"
        % (number_of_lines[0], bucket, prefix)
    )


def read_s3_jsonl_as_df(
    bucket: str,
    prefix: str,
) -> pd.DataFrame:
    _assert_s3_enabled()
    s3 = S3FileSystem()
    with s3.open(
        ParsedS3Path(bucket=bucket, prefix=prefix).stripped_path,
        mode="rb",
    ) as f:
        df = pd.read_json(f, lines=True)
    return df


def read_lines_s3(
    bucket: str,
    prefix: str,
    lazy: bool = True,
) -> Iterable[str]:
    _assert_s3_enabled()
    if lazy:
        return _read_lines_s3_lazy(bucket, prefix)
    else:
        return _read_lines_s3_greedy(bucket, prefix)


def _read_lines_s3_greedy(
    bucket: str,
    prefix: str,
) -> List[str]:
    s3 = S3FileSystem()
    with s3.open(
        ParsedS3Path(bucket=bucket, prefix=prefix).stripped_path,
        mode="rb",
    ) as f:
        lines = f.readlines()
    return lines


def _read_lines_s3_lazy(
    bucket: str,
    prefix: str,
) -> Iterator[str]:
    s3 = S3FileSystem()
    with s3.open(
        ParsedS3Path(bucket=bucket, prefix=prefix).stripped_path,
        mode="rb",
    ) as f:
        try:
            yield f.readline()
        except Exception:
            raise StopIteration



from contextlib import nullcontext
from typing import Any, Callable, ContextManager, Iterable, Iterator, Sequence, TypeVar

import pandas as pd
import pytest

from elabftwcontrol.core.interfaces import HasIDAndMetadata
from elabftwcontrol.core.metadata import MetadataField
from elabftwcontrol.core.transformers import (
    CSVElabDataTableTransformer,
    CSVLongMetadataTableTransformer,
    JSONTransformer,
    LazyWideTableUtils,
    MultiPandasDataFrameTransformer,
    ObjectTypes,
    SplitDataFrame,
)

T = TypeVar("T")


class MockDictable:
    def __init__(self, data: dict[str, Any]) -> None:
        self.data = data

    def to_dict(self) -> dict[str, Any]:
        return self.data


class TestJSONTransformer:
    def test_obj_json_transform(self) -> None:
        data = {"a": 1, "b": {"c": [1, 2, 3], "d": "test"}}
        objs = [MockDictable(data)]
        result = list(JSONTransformer.new(indent=None)(objs))
        expected = ['{"a": 1, "b": {"c": [1, 2, 3], "d": "test"}}']
        assert result == expected

    def test_json_transformer(self) -> None:
        n_obj = 4

        def create_data(a: int) -> dict[str, Any]:
            return {"a": a, "b": {"c": [1, 2, 3], "d": "test"}}

        objs = [MockDictable(create_data(a)) for a in range(n_obj)]

        transformer = JSONTransformer.new(indent=None)
        transformed = transformer(objs)

        for i, result in enumerate(transformed):
            expected = '{"a": ' + f"{i}" + ', "b": {"c": [1, 2, 3], "d": "test"}}'
            assert expected == result


class MockIdAndMetadata(HasIDAndMetadata):
    def __init__(self, metadata: str) -> None:
        self.id = 1
        self.metadata = metadata


class TestMultiPandasDataFrameTransformer:
    def verify_result(
        self,
        result: Iterable[SplitDataFrame],
        expected: Iterable[SplitDataFrame],
    ) -> None:
        for r, e in zip(result, expected):
            assert r.key == e.key
            pd.testing.assert_frame_equal(r.data, e.data)

    def test_transformer(self) -> None:
        def splitter(raw_data: pd.DataFrame) -> Iterator[SplitDataFrame]:
            yield SplitDataFrame(
                key="key1",
                data=pd.DataFrame({"a": [1, 2, 3], "b": ["x", "y", "z"]}),
            )
            yield SplitDataFrame(
                key="key2",
                data=pd.DataFrame({"a": [4, 5, 6], "b": ["a", "b", "c"]}),
            )

        def df_transform_getter(key: str) -> Callable[[pd.DataFrame], pd.DataFrame]:
            if key == "key1":

                def transformer(data: pd.DataFrame) -> pd.DataFrame:
                    return data[data["a"] < 3].reset_index(drop=True)

            elif key == "key2":

                def transformer(data: pd.DataFrame) -> pd.DataFrame:
                    return data[data["b"] > "a"].reset_index(drop=True)

            else:
                raise ValueError

            return transformer

        transformer = MultiPandasDataFrameTransformer(
            splitter=splitter,
            df_transform_getter=df_transform_getter,
        )
        result = transformer(pd.DataFrame())
        expected = (
            SplitDataFrame(
                key="key1",
                data=pd.DataFrame({"a": [1, 2], "b": ["x", "y"]}),
            ),
            SplitDataFrame(
                key="key2",
                data=pd.DataFrame({"a": [5, 6], "b": ["b", "c"]}),
            ),
        )
        self.verify_result(result, expected)

    @pytest.mark.parametrize(
        ["input", "obj_type", "categories", "expected", "context"],
        (
            (
                pd.DataFrame(
                    {
                        "category_title": ["a", "b", "a", "a"],
                        "other_column": [1, 2, 3, 4],
                    }
                ),
                ObjectTypes.ITEM,
                None,
                (
                    SplitDataFrame(
                        key="a",
                        data=pd.DataFrame(
                            {
                                "category_title": ["a", "a", "a"],
                                "other_column": [1, 3, 4],
                            }
                        ),
                    ),
                    SplitDataFrame(
                        key="b",
                        data=pd.DataFrame(
                            {
                                "category_title": ["b"],
                                "other_column": [2],
                            }
                        ),
                    ),
                ),
                nullcontext(),
            ),
            (
                pd.DataFrame(
                    {
                        "title": ["a 1", "b 2", "a 3", "a 4"],
                        "other_column": [1, 2, 3, 4],
                    }
                ),
                ObjectTypes.EXPERIMENT,
                ("a", "b"),
                (
                    SplitDataFrame(
                        key="a",
                        data=pd.DataFrame(
                            {
                                "title": ["a 1", "a 3", "a 4"],
                                "other_column": [1, 3, 4],
                            }
                        ),
                    ),
                    SplitDataFrame(
                        key="b",
                        data=pd.DataFrame(
                            {
                                "title": ["b 2"],
                                "other_column": [2],
                            }
                        ),
                    ),
                ),
                nullcontext(),
            ),
            (
                pd.DataFrame(
                    {
                        "title": ["a 1", "b 2", "a 3", "a 4"],
                        "other_column": [1, 2, 3, 4],
                    }
                ),
                ObjectTypes.EXPERIMENT,
                None,
                None,
                pytest.raises(ValueError),
            ),
            (
                pd.DataFrame(
                    {
                        "title": ["a 1", "b 2", "a 3", "a 4"],
                        "other_column": [1, 2, 3, 4],
                    }
                ),
                ObjectTypes.EXPERIMENTS_TEMPLATE,
                None,
                (
                    SplitDataFrame(
                        key="experiments_template",
                        data=pd.DataFrame(
                            {
                                "title": ["a 1", "b 2", "a 3", "a 4"],
                                "other_column": [1, 2, 3, 4],
                            }
                        ),
                    ),
                ),
                nullcontext(),
            ),
        ),
    )
    def test_for_raw_tables(
        self,
        input: pd.DataFrame,
        obj_type: ObjectTypes,
        categories: Sequence[str] | None,
        expected: Iterable[SplitDataFrame],
        context: ContextManager,
    ) -> None:
        with context:
            transformer = MultiPandasDataFrameTransformer.for_raw_tables(
                object_type=obj_type,
                categories=categories,
            )
            result = transformer(input)
            self.verify_result(result, expected)


class TestCSVLongMetadataTableTransformer:
    def test_csv_long_metadata_transformer(self) -> None:
        metadata = """\
{
    "elabftw": {
        "extra_fields_groups": [
            {"id": 2, "name": "somegroup"}
        ]
    },
    "extra_fields": {
        "field1": {},
        "field2": {
            "group_id": "1",
            "position": 10,
            "value": "blabla",
            "unit": "kg"
        },
        "field3": {
            "type": "number",
            "group_id": "2",
            "position": 5,
            "value": "1234",
            "unit": ""
        }
    }
}
"""
        obj = MockIdAndMetadata(metadata)
        expected = [
            {
                "_id": 1,
                "field_name": "field1",
                "field_type": "text",
                "field_group": "",
                "field_position": -1,
                "field_description": "",
                "field_value": "",
                "field_unit": "",
            },
            {
                "_id": 1,
                "field_name": "field2",
                "field_type": "text",
                "field_group": "",
                "field_position": 10,
                "field_description": "",
                "field_value": "blabla",
                "field_unit": "kg",
            },
            {
                "_id": 1,
                "field_name": "field3",
                "field_type": "number",
                "field_group": "somegroup",
                "field_position": 5,
                "field_description": "",
                "field_value": "1234",
                "field_unit": "",
            },
        ]

        transformer = CSVLongMetadataTableTransformer.new()
        result = transformer([obj, obj])
        for e, r in zip(result, expected * 2):
            assert e == r


class MockGenericObject:
    def __init__(self, a: int, b: str, c: bool) -> None:
        self.a = a
        self.b = b
        self.c = c


class TestCSVElabDAtaTableTransformer:
    def test_csv_elab_data_transformer(self) -> None:
        def column_name_sanitizer(col: str) -> str:
            if col == "a":
                return "z"
            else:
                return col

        obj = MockGenericObject(a=1, b="2", c=True)
        columns = ["a", "b", "d"]

        transformer = CSVElabDataTableTransformer(
            columns=columns,
            column_name_sanitizer=column_name_sanitizer,
        )
        result = transformer.transform(obj)
        expected = {"z": 1, "b": "2"}

        assert expected == result


class MockGenericObjectIdAndMetadata:
    def __init__(
        self,
        a: int,
        b: str,
        c: bool,
        fields: list[MetadataField],
    ) -> None:
        self.id = 1
        self.a = a
        self.b = b
        self.c = c
        self.metadata = fields
        self.fields = fields


def dummy_field_to_cell_value(field: MetadataField) -> str:
    return str(field.value) + "_test"


def dummy_column_name_sanitizer(col: str) -> str:
    if col == "a":
        return "z"
    elif col == "f1":
        return "zz1"
    else:
        return col


class TestWideTableUtils:
    obj_columns = ["a", "b", "d"]
    meta_columns = ["f1", "f2", "f4"]

    @pytest.mark.parametrize(
        ("sanitizer", "expected"),
        (
            (dummy_column_name_sanitizer, ["_a", "_b", "_d", "zz1", "f2", "f4"]),
            (None, ["_a", "_b", "_d", "f1", "f2", "f4"]),
        ),
    )
    def test_create_header(
        self,
        sanitizer: Callable[[str], str],
        expected: Sequence[str],
    ) -> None:
        result = LazyWideTableUtils.create_header(
            object_columns=self.obj_columns,
            metadata_columns=self.meta_columns,
            column_name_sanitizer=sanitizer,
        )
        assert result == expected



from __future__ import annotations

import os
from functools import cached_property
from pathlib import Path
from typing import (
    Any,
    Collection,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Protocol,
    Sequence,
    TypeVar,
    Union,
)

from elabapi_python import (  # type: ignore
    ApiClient,
    Experiment,
    ExperimentsApi,
    ExperimentsTemplatesApi,
    ExperimentTemplate,
    InfoApi,
    Item,
    ItemsApi,
    ItemsBody,
    ItemsType,
    ItemsTypesApi,
    Link,
    LinksToExperimentsApi,
    LinksToItemsApi,
    Tag,
    TagsApi,
    Upload,
    UploadsApi,
)

from elabftwcontrol._logging import logger
from elabftwcontrol.configure import AccessConfig, MultiConfig
from elabftwcontrol.defaults import DEFAULT_CONFIG_FILE
from elabftwcontrol.types import EntityTypes, SingleObjectTypes

Pathlike = Union[Path, str]


DEFAULT_REQUEST_BATCH_SIZE = 1000


class ElabftwApi:
    def __init__(
        self,
        client: ApiClient,
        request_batch_size: int = DEFAULT_REQUEST_BATCH_SIZE,
    ) -> None:
        self.client = client
        self.request_batch_size = request_batch_size

    @property
    def host_name(self) -> str:
        return self.client.configuration.host

    @cached_property
    def info(self) -> InfoApi:
        return InfoApi(self.client)

    @cached_property
    def items(self) -> ItemCRUD:
        return ItemCRUD.from_client(self.client)

    @cached_property
    def experiments(self) -> ExperimentCRUD:
        return ExperimentCRUD.from_client(self.client)

    @cached_property
    def items_types(self) -> ItemsTypeCRUD:
        return ItemsTypeCRUD.from_client(self.client)

    @cached_property
    def experiments_templates(self) -> ExperimentTemplateCRUD:
        return ExperimentTemplateCRUD.from_client(self.client)

    @cached_property
    def uploads(self) -> UploadsCRUD:
        return UploadsCRUD.from_client(self.client)

    @cached_property
    def tags(self) -> TagCRUD:
        return TagCRUD.from_client(self.client)

    @cached_property
    def links(self) -> LinkCRUD:
        return LinkCRUD.from_client(self.client)

    @classmethod
    def from_config_file(
        cls,
        filepath: Pathlike = DEFAULT_CONFIG_FILE,
        profile: str = "default",
    ) -> ElabftwApi:
        """Create an eLabFTW client from a config file and profile"""
        configuration = MultiConfig.from_file(filepath).get_profile(profile)
        return cls._create(configuration)

    @classmethod
    def new(
        cls,
        host_url: Optional[str] = None,
        api_key: Optional[str] = None,
        verify_ssl: Optional[bool] = None,
        debug: Optional[bool] = None,
    ) -> ElabftwApi:
        host_url = host_url or os.getenv("ELABCTL_HOST_URL")
        if not host_url:
            raise ValueError(
                "Host URL must be specified or defined in the ELABCTL_HOST_URL environment variable."
            )

        api_key = api_key or os.getenv("ELABCTL_API_KEY")
        if not api_key:
            raise ValueError(
                "API key must be specified or defined in the ELABCTL_API_KEY environment variable."
            )

        if verify_ssl is None:
            verify_ssl = bool(int(os.getenv("ELABCTL_VERIFY_SSL", "1")))

        if debug is None:
            debug = bool(int(os.getenv("ELABCTL_DEBUG", "0")))

        configuration = AccessConfig(
            host_url=host_url,
            api_key=api_key,
            verify_ssl=verify_ssl,
            debug=debug,
        )
        return cls._create(configuration)

    @classmethod
    def _create(
        cls,
        configuration: AccessConfig,
    ) -> ElabftwApi:
        api_client = ApiClient(configuration.get_api_config())
        api_client.set_default_header(
            header_name="Authorization",
            header_value=configuration.api_key,
        )
        return cls(client=api_client)


class ApiResponseObject(Protocol):
    id: int

    def to_dict(self) -> dict[str, Any]: ...


T = TypeVar("T", bound=ApiResponseObject, covariant=True)


class EntityRUD(Protocol[T]):
    def read(self, id: int) -> T: ...

    def iter(self) -> Iterable[T]: ...

    def patch(self, id: int, body: dict[str, Any]) -> None: ...

    def delete(self, id: int) -> None: ...


class GroupEntityCreate(Protocol):
    def create(self) -> int: ...


class SingleEntityCreate(Protocol):
    def create(self, category_id: int) -> int: ...


class ItemsTypeCRUD(EntityRUD, GroupEntityCreate):
    def __init__(
        self,
        api: ItemsTypesApi,
    ) -> None:
        self.api = api

    @classmethod
    def from_client(
        cls,
        api_client: ApiClient,
    ) -> ItemsTypeCRUD:
        return cls(api=ItemsTypesApi(api_client))

    def create(self) -> int:
        _, _, response = self.api.post_items_types_with_http_info()
        category_id = self._get_new_category_id(response)
        logger.info(f"Created new category: {category_id}")
        return category_id

    def get_by_name(self, name: str) -> ItemsType:
        category = None
        templates = self.api.read_items_types()
        for template in templates:
            if template.title == name:
                category = template
                break
        else:
            raise ValueError(f"Category '{name}' was not found!")
        return self.read(category.id)

    def search(self, q: str) -> List[ItemsType]:
        templates = self.api.read_items_types()

        if not q:
            return templates

        to_return: List[ItemsType] = []
        for template in templates:
            if template.title is not None and q in template.title:
                to_return.append(template)
            if template.body is not None and q in template.body:
                to_return.append(template)

        return to_return

    def read(self, id: int) -> ItemsType:
        return self.api.get_items_type(id)

    def iter(self, ids: Optional[Sequence[int]] = None) -> List[ItemsType]:
        # it is much more efficient to make one request than one per id
        all_types = self.api.read_items_types()
        if not ids:
            return all_types
        item_type_map = {item_type.id: item_type for item_type in all_types}
        return [item_type_map[id] for id in ids]

    def iter_full(
        self,
        category_ids: Optional[Iterable[int]] = None,
    ) -> List[ItemsType]:
        if not category_ids:
            templates = self.iter()
            category_ids = (template.id for template in templates)

        threads = []
        for category_id in category_ids:
            threads.append(self.api.get_items_type(category_id, async_req=True))
        return [thread.get() for thread in threads]

    def patch(
        self,
        id: int,
        body: Dict[str, Any],
    ) -> None:
        logger.debug(f"Patch body: {body}")
        self.api.patch_items_type(id, body=ItemsType(**body))
        logger.info(f"Patched category {id}")

    def delete(
        self,
        id: int,
    ) -> None:
        self.api.delete_items_type(id)
        logger.info(f"Deleted category {id}")

    def _get_new_category_id(self, response: Any) -> int:
        return int(response.get("Location").split("=")[-1])


class ExperimentTemplateCRUD(EntityRUD, GroupEntityCreate):
    def __init__(self, api: ExperimentsTemplatesApi) -> None:
        self.api = api

    @classmethod
    def from_client(cls, api_client: ApiClient) -> ExperimentTemplateCRUD:
        return cls(api=ExperimentsTemplatesApi(api_client))

    def create(self) -> int:
        _, _, response = self.api.post_experiment_template_with_http_info()
        template_id = self._get_new_template_id(response)
        logger.info(f"Created new template: {template_id}")
        return template_id

    def search(self, q: str) -> List[ExperimentTemplate]:
        templates = self.api.read_experiments_templates()

        if not q:
            return templates

        to_return: List[ExperimentTemplate] = []
        for template in templates:
            if template.title is not None and q in template.title:
                to_return.append(template)
            if template.body is not None and q in template.body:
                to_return.append(template)

        return to_return

    def read(self, id: int) -> ExperimentTemplate:
        return self.api.get_experiment_template(id)

    def iter(self, ids: Optional[Sequence[int]] = None) -> List[ExperimentTemplate]:
        # it is much more efficient to make one request than one per id
        all_templates = self.api.read_experiments_templates()
        if not ids:
            return all_templates
        template_map = {template.id: template for template in all_templates}
        return [template_map[id] for id in ids]

    def iter_full(
        self,
        ids: Optional[Iterable[int]] = None,
    ) -> List[ExperimentTemplate]:
        if not ids:
            templates = self.iter()
            ids = (template.id for template in templates)

        threads = []
        for category_id in ids:
            threads.append(
                self.api.get_experiment_template(category_id, async_req=True)
            )
        return [thread.get() for thread in threads]

    def patch(self, id: int, body: Dict[str, Any]) -> None:
        self.api.patch_experiment_template(id, body=body)
        logger.info(f"Patched template {id}")

    def delete(self, id: int) -> None:
        self.api.delete_experiment_template(id)
        logger.info(f"Deleted template {id}")

    def _get_new_template_id(self, response: Any) -> int:
        return int(response.get("Location").split("/")[-1])


class ItemCRUD(EntityRUD, SingleEntityCreate):
    def __init__(
        self,
        api: ItemsApi,
        read_batch_size: int,
    ) -> None:
        self.api = api
        self.read_batch_size = read_batch_size

    @classmethod
    def from_client(
        cls,
        api_client: ApiClient,
        read_batch_size: int = DEFAULT_REQUEST_BATCH_SIZE,
    ) -> ItemCRUD:
        return cls(
            api=ItemsApi(api_client),
            read_batch_size=read_batch_size,
        )

    def create(self, category_id: int) -> int:
        _, _, response = self.api.post_item_with_http_info(
            body=ItemsBody(category_id=category_id)
        )
        new_item_id = self._get_new_item_id(response)
        logger.info(f"Created item {new_item_id}")
        return new_item_id

    def read(self, id: int) -> Item:
        return self.api.get_item(id)

    def search(self, q: str, **kwargs: Any) -> List[Item]:
        return self.api.read_items(q=q, **kwargs)

    def iter(
        self,
        ids: Optional[Collection[int]] = None,
        **kwargs: Any,
    ) -> Iterator[Item]:
        """Return items by making batched requests"""
        read_number_of_items = 0
        while True:
            items = self.api.read_items(
                limit=self.read_batch_size - 1,
                offset=read_number_of_items,
                **kwargs,
            )
            if not items:
                logger.debug("No more items found, exiting.")
                break

            for item in items:
                if ids and item.id not in ids:
                    continue

                yield item

            if len(items) < self.read_batch_size:
                logger.debug("Less items found than the batch size. Exiting.")
                break

            read_number_of_items += self.read_batch_size

    def iter_full(
        self,
        ids: Optional[Iterable[int]] = None,
        **kwargs: Any,
    ) -> Iterator[Item]:
        """Request all items individually to fetch all information"""
        if not ids:
            items = self.iter()
            ids = (item.id for item in items)

        threads = []
        for item_id in ids:
            threads.append(self.api.get_item(item_id, async_req=True, **kwargs))

        for thread in threads:
            yield thread.get()

    def patch(self, id: int, body: Dict[str, Any]) -> None:
        self.api.patch_item(id, body=body)
        logger.info(f"Patched item {id}")

    def delete(self, id: int) -> None:
        self.api.delete_item(id)
        logger.info(f"Deleted item {id}")

    def _get_new_item_id(self, response: Any) -> int:
        return int(response.get("Location").split("/")[-1])


class ExperimentCRUD(EntityRUD, SingleEntityCreate):
    def __init__(
        self,
        api: ExperimentsApi,
        read_batch_size: int,
    ) -> None:
        self.api = api
        self.read_batch_size = read_batch_size

    @classmethod
    def from_client(
        cls,
        api_client: ApiClient,
        read_batch_size: int = DEFAULT_REQUEST_BATCH_SIZE,
    ) -> ExperimentCRUD:
        return cls(
            api=ExperimentsApi(api_client),
            read_batch_size=read_batch_size,
        )

    def create(self, category_id: int = -1) -> int:
        _, _, response = self.api.post_experiment_with_http_info(
            body={"category_id": category_id}
        )
        new_experiment_id = self._get_new_experiment_id(response)
        logger.info(f"Created new experiment {new_experiment_id}")
        return new_experiment_id

    def read(self, id: int) -> Experiment:
        return self.api.get_experiment(id)

    def search(self, q: str, **kwargs: Any) -> List[Experiment]:
        return self.api.read_experiments(q=q, **kwargs)

    def iter(
        self,
        ids: Optional[Collection[int]] = None,
        **kwargs: Any,
    ) -> Iterator[Experiment]:
        """Return items by making batched requests"""
        read_number_of_experiments = 0
        while True:
            experiments = self.api.read_experiments(
                limit=self.read_batch_size - 1,
                offset=read_number_of_experiments,
                **kwargs,
            )
            if not experiments:
                logger.debug("No more experiments found, exiting.")
                break

            for experiment in experiments:
                if ids and experiment.id not in ids:
                    continue

                yield experiment

            if len(experiments) < self.read_batch_size:
                logger.debug("Less experiments found than the batch size. Exiting.")
                break

            read_number_of_experiments += self.read_batch_size

    def iter_full(
        self,
        ids: Optional[Iterable[int]] = None,
        **kwargs: Any,
    ) -> Iterator[Experiment]:
        """Request all experiments individually to fetch all information"""
        if not ids:
            experiments = self.iter()
            ids = (experiment.id for experiment in experiments)

        threads = []
        for experiment_id in ids:
            threads.append(
                self.api.get_experiment(experiment_id, async_req=True, **kwargs)
            )

        for thread in threads:
            yield thread.get()

    def patch(self, id: int, body: Dict[str, Any]) -> None:
        self.api.patch_experiment(id, body=body)
        logger.info(f"Patched experiment {id}")

    def delete(self, id: int) -> None:
        self.api.delete_experiment(id)
        logger.info(f"Deleted experiment {id}")

    def _get_new_experiment_id(self, response: Any) -> int:
        return int(response.get("Location").split("/")[-1])


class UploadsCRUD:
    def __init__(
        self,
        api: UploadsApi,
    ) -> None:
        self.api = api

    @classmethod
    def from_client(cls, api_client: ApiClient) -> UploadsCRUD:
        return cls(api=UploadsApi(api_client))

    def iter(self, entity_type: SingleObjectTypes, id: int) -> list[Upload]:
        return self.api.read_uploads(entity_type.value, id)

    def read(
        self,
        entity_type: SingleObjectTypes,
        id: int,
        sub_id: int,
    ) -> bytes:
        return self.api.read_upload(
            entity_type=entity_type.value,
            id=id,
            subid=sub_id,
            format="binary",
            _preload_content=False,
        )

    def patch(
        self,
        entity_type: SingleObjectTypes,
        id: int,
        sub_id: int,
        file: bytes,
        comment: str,
    ) -> None:
        self.api.post_upload_replace(
            entity_type=entity_type.value,
            id=id,
            subid=sub_id,
            file=file,
            comment=comment,
        )

    def create(
        self,
        entity_type: SingleObjectTypes,
        id: int,
        file: bytes,
        comment: str,
    ) -> None:
        self.api.post_upload(
            entity_type=entity_type.value,
            id=id,
            file=file,
            comment=comment,
        )

    def delete(
        self,
        entity_type: SingleObjectTypes,
        id: int,
        sub_id: int,
    ) -> None:
        self.api.delete_upload(
            entity_type=entity_type.value,
            id=id,
            subid=sub_id,
        )


class LinkCRUD:
    def __init__(
        self,
        links_to_items_api: LinksToItemsApi,
        links_to_experiments_api: LinksToExperimentsApi,
    ) -> None:
        self.links_to_items_api = links_to_items_api
        self.links_to_experiments_api = links_to_experiments_api

    @classmethod
    def from_client(cls, api_client: ApiClient) -> LinkCRUD:
        return cls(
            links_to_items_api=LinksToItemsApi(api_client),
            links_to_experiments_api=LinksToExperimentsApi(api_client),
        )

    def create(
        self,
        source_type: SingleObjectTypes,
        source_id: int,
        destination_type: SingleObjectTypes,
        destination_id: int,
    ) -> None:
        self._verify_source_and_destination_types(source_type, destination_type)

        if destination_type == "items":
            method = self.links_to_items_api.post_entity_items_links
        else:
            method = self.links_to_experiments_api.post_entity_experiments_links

        method(
            entity_type=source_type,
            id=source_id,
            subid=destination_id,
            body={"action": "create"},
        )

    def exists(
        self,
        source_type: SingleObjectTypes,
        source_id: int,
        destination_type: SingleObjectTypes,
        destination_id: int,
    ) -> bool:
        self._verify_source_and_destination_types(source_type, destination_type)

        links = self.read(
            source_type=source_type,
            source_id=source_id,
            destination_type=destination_type,
        )
        for link in links:
            if link.itemid == destination_id:
                return True
        return False

    def read(
        self,
        source_type: SingleObjectTypes,
        source_id: int,
        destination_type: SingleObjectTypes,
    ) -> List[Link]:
        self._verify_source_and_destination_types(source_type, destination_type)

        if destination_type == "items":
            method = self.links_to_items_api.read_entity_items_links
        else:
            method = self.links_to_experiments_api.read_entity_experiments_links

        return method(
            entity_type=source_type,
            id=source_id,
        )

    def delete(
        self,
        source_type: SingleObjectTypes,
        source_id: int,
        destination_type: SingleObjectTypes,
        destination_id: int,
    ) -> None:
        self._verify_source_and_destination_types(source_type, destination_type)

        if destination_type == "items":
            # possible typo in eLabFTW API spec
            method = self.links_to_items_api.delete_entitiy_items_link
        else:
            method = self.links_to_experiments_api.delete_entity_experiments_link

        method(
            entity_type=source_type,
            id=source_id,
            subid=destination_id,
        )

    @classmethod
    def _verify_source_and_destination_types(
        cls,
        source_type: str,
        destination_type: str,
    ) -> None:
        assert source_type in ("items", "experiments")
        assert destination_type in ("items", "experiments")


class TagCRUD:
    def __init__(
        self,
        api: TagsApi,
    ) -> None:
        self.api = api

    @classmethod
    def from_client(cls, api_client: ApiClient) -> TagCRUD:
        return cls(api=TagsApi(api_client))

    def create(
        self,
        entity_type: EntityTypes,
        entity_id: int,
        tag: str,
    ) -> None:
        self.api.post_tag(
            entity_type.value,
            entity_id,
            body={"tag": tag},
        )
        logger.debug(f"Added tag '{tag}' to {entity_type} {entity_id}")

    def read(
        self,
        entity_type: EntityTypes,
        entity_id: int,
        tag_id: int,
    ) -> Tag:
        return self.api.read_tag(entity_type.value, entity_id, tag_id)

    def iter(
        self,
        entity_type: EntityTypes,
        entity_id: int,
    ) -> List[Tag]:
        return self.api.read_tags(entity_type.value, entity_id)

    def delete(
        self,
        entity_type: EntityTypes,
        entity_id: int,
        tag_id: int,
    ) -> None:
        self.api.patch_tag(
            entity_type.value,
            entity_id,
            tag_id,
            body={"action": "unreference"},
        )
        logger.debug(f"Removed tag {tag_id} from {entity_type.value} {entity_id}")

    def delete_all(
        self,
        entity_type: EntityTypes,
        entity_id: int,
    ) -> None:
        self.api.delete_tag(entity_type.value, entity_id)
        logger.debug(f"Deleted all tags from {entity_type.value} {entity_id}")



from __future__ import annotations

from itertools import chain
from pathlib import Path
from typing import Any, Dict, Iterable, List, Mapping, NamedTuple, Sequence, Union

from elabftwcontrol._logging import logger
from elabftwcontrol.core.manifests import ElabObjManifests, ManifestIndex
from elabftwcontrol.utils import read_yaml

Pathlike = Union[Path, str]


class Manifests(NamedTuple):
    manifests: Sequence[Mapping[str, Any]]


class ManifestParser:
    """Parses desired state manifests from a predefined file structure"""

    def __init__(
        self,
        path: Pathlike,
    ) -> None:
        self.path = Path(path)

    def parse(self) -> ManifestIndex:
        files = self.get_files(self.path)
        raw_object_definitions = self.parse_definition_files(files)
        manifests = self.interpret_definitions(raw_object_definitions)
        return ManifestIndex.from_manifests(manifests.manifests)

    @classmethod
    def get_files(cls, path: Path) -> list[Path]:
        if path.is_dir():
            return cls.find_yml_files_in_folder(path)
        else:
            return [path]

    @classmethod
    def find_yml_files_in_folder(
        cls,
        definitions_folder: Path,
    ) -> list[Path]:
        definition_files = list(
            map(
                Path,
                chain(
                    definitions_folder.rglob("*.yaml"),
                    definitions_folder.rglob("*.yml"),
                ),
            )
        )
        logger.info(f"Found definition files: {definition_files}")
        return definition_files

    @classmethod
    def parse_definition_files(
        cls,
        definition_files: Iterable[Pathlike],
    ) -> Manifests:
        object_definitions = []
        for filepath in definition_files:
            parsed_manifests = cls.parse_definition_file(filepath)
            if not parsed_manifests:
                continue
            for parsed_manifest in parsed_manifests:
                logger.debug(parsed_manifest)
                object_definitions.append(parsed_manifest)

        return Manifests(manifests=object_definitions)

    @classmethod
    def parse_definition_file(
        cls,
        filepath: Pathlike,
    ) -> List[Dict[str, Any]]:
        logger.info(f"Parsing definition file: {filepath}")
        data = read_yaml(filepath)

        if isinstance(data, dict):
            data = [data]

        return data

    @classmethod
    def interpret_definitions(
        cls,
        raw_definitions: Manifests,
    ) -> ElabObjManifests:
        return ElabObjManifests(**raw_definitions._asdict())



from __future__ import annotations

import json
from dataclasses import dataclass
from typing import Any, Callable, Iterable, Iterator, Mapping

from pydantic import BaseModel
from typing_extensions import Self

from elabftwcontrol._logging import logger
from elabftwcontrol.client import ElabftwApi
from elabftwcontrol.core.interfaces import ElabApiObjInterface, Pathlike
from elabftwcontrol.core.models import CLASS_TO_OBJTYPE, OBJTYPE_TO_CLASS, IdNode
from elabftwcontrol.core.models import ObjectTypes as ElabObjectType
from elabftwcontrol.core.parsers import MetadataModel, MetadataParser
from elabftwcontrol.utils import parse_tag_id_str, parse_tag_str


@dataclass(frozen=True)
class TypedObj:
    type: ElabObjectType
    obj: ElabApiObjInterface


class SerializedObj(BaseModel):
    """For writing and reading obj to JSON file"""

    type: str
    data: dict[str, Any]


class SerializedState(BaseModel):
    """For writing and reading state to JSON file"""

    objects: list[SerializedObj]


def get_tags(obj: Any) -> list[str]:
    """Type-unsafe hack to extract tags from objects"""
    parsed = None
    if hasattr(obj, "tags"):
        parsed = parse_tag_str(obj.tags)
    if parsed is None:
        parsed = []
    return parsed


def get_tag_ids(obj: Any) -> list[int]:
    """Type-unsafe hack to extract tag ids from objects"""
    parsed = None
    if hasattr(obj, "tag_ids"):
        parsed = parse_tag_id_str(obj.tag_ids)
    if parsed is None:
        parsed = []
    return parsed


@dataclass(frozen=True)
class EnrichedObj:
    type: ElabObjectType
    obj: ElabApiObjInterface
    metadata: MetadataModel

    @property
    def tags(self) -> list[str]:
        return get_tags(self.obj)

    @property
    def tag_ids(self) -> list[int]:
        return get_tag_ids(self.obj)

    @property
    def tag_map(self) -> dict[str, int]:
        return {tag: tag_id for tag, tag_id in zip(self.tags, self.tag_ids)}

    @classmethod
    def from_typed_obj(
        cls,
        typed_obj: TypedObj,
        metadata_parser: Callable[[str | None], MetadataModel],
    ) -> Self:
        parsed_metadata = metadata_parser(typed_obj.obj.metadata)
        return cls(
            type=typed_obj.type,
            obj=typed_obj.obj,
            metadata=parsed_metadata,
        )

    @property
    def raw_dict(self) -> dict[str, Any]:
        return self.obj.to_dict()

    def get_values(self, keys: Iterable[str]) -> dict[str, Any]:
        raw_dict = self.raw_dict
        return {key: raw_dict[key] for key in keys}

    @property
    def id(self) -> int:
        return self.obj.id

    def to_typed_obj(self) -> TypedObj:
        return TypedObj(type=self.type, obj=self.obj)

    def to_dict(self) -> dict[str, Any]:
        return self.to_serialized().model_dump(mode="json")

    @classmethod
    def from_dict(
        cls,
        data: dict[str, Any],
        metadata_parser: Callable[[str | None], MetadataModel],
    ) -> Self:
        serialized = SerializedObj(**data)
        return cls.from_serialized(serialized, metadata_parser)

    def to_serialized(self) -> SerializedObj:
        return SerializedObj(
            type=self.type.value,
            data=self.obj.to_dict(),
        )

    @classmethod
    def from_serialized(
        cls,
        serialized: SerializedObj,
        metadata_parser: Callable[[str | None], MetadataModel],
    ) -> Self:
        obj_type = ElabObjectType(serialized.type)
        obj_class = OBJTYPE_TO_CLASS[obj_type]
        obj = obj_class(**serialized.data)
        parsed_metadata = metadata_parser(obj.metadata)
        return cls(
            type=obj_type,
            obj=obj,
            metadata=parsed_metadata,
        )

    @property
    def name(self) -> str:
        elabctl_data = self.metadata.elabftwcontrol
        if elabctl_data is not None:
            if self.type.is_template():
                name = elabctl_data.template_name
            else:
                name = elabctl_data.name
        else:
            name = None

        return name or self._default_name

    @property
    def _default_name(self) -> str:
        return f"{self.type.value}_{self.id}"


@dataclass(frozen=True)
class State:
    """The current state of eLab managed by elabftwcontrol"""

    elab_obj: Mapping[IdNode, EnrichedObj]

    def items(self) -> Iterator[tuple[IdNode, EnrichedObj]]:
        return iter(self.elab_obj.items())

    def __getitem__(self, id_node: IdNode) -> EnrichedObj:
        return self.elab_obj[id_node]

    def get(self, id_node: IdNode) -> EnrichedObj | None:
        return self.elab_obj.get(id_node)

    def __contains__(self, id_node: IdNode) -> bool:
        return id_node in self.elab_obj

    def __len__(self) -> int:
        return len(self.elab_obj)

    @classmethod
    def empty(cls) -> Self:
        return cls({})

    @classmethod
    def from_api(
        cls,
        api: ElabftwApi,
        skip_untracked: bool = True,
    ) -> Self:
        """Get the state from a remote eLabFTW instance via the API"""
        all_obj = cls._pull(api)
        return cls._from_typed_api_objs(all_obj, skip_untracked=skip_untracked)

    @classmethod
    def from_api_objs(
        cls,
        api_objs: Iterable[ElabApiObjInterface],
    ) -> Self:
        typed_objs = cls._classify_objects(api_objs)
        return cls._from_typed_api_objs(typed_objs, skip_untracked=False)

    def to_serialized(self) -> SerializedState:
        return SerializedState(
            objects=[obj.to_serialized() for obj in self.elab_obj.values()]
        )

    @classmethod
    def from_serialized(cls, serialized: SerializedState) -> Self:
        meta_parser = MetadataParser()
        api_objs = (
            EnrichedObj.from_serialized(obj, metadata_parser=meta_parser)
            for obj in serialized.objects
        )
        return cls._from_enriched_objs(api_objs)

    def to_file(self, filepath: Pathlike) -> None:
        serialized = self.to_serialized()
        with open(filepath, "w") as f:
            f.write(serialized.model_dump_json(indent=2))

    @classmethod
    def from_file(cls, filepath: Pathlike) -> Self:
        with open(filepath, "r") as f:
            data = json.load(f)
        return cls.from_serialized(SerializedState(**data))

    @classmethod
    def _from_typed_api_objs(
        cls,
        typed_objs: Iterator[TypedObj],
        skip_untracked: bool,
    ) -> Self:
        meta_parser = MetadataParser()
        enriched_objs = cls._enrich_objs(
            typed_objs,
            metadata_parser=meta_parser,
            skip_untracked=skip_untracked,
        )
        return cls._from_enriched_objs(enriched_objs)

    @classmethod
    def _from_enriched_objs(
        cls,
        enriched_objs: Iterable[EnrichedObj],
    ) -> Self:
        elab_obj = {
            IdNode(enriched_obj.type, enriched_obj.id): enriched_obj
            for enriched_obj in enriched_objs
        }
        return cls(elab_obj=elab_obj)

    @classmethod
    def _pull(cls, api: ElabftwApi) -> Iterator[TypedObj]:
        """Pull all relevant objects from the API"""
        fetch_jobs: list[
            tuple[ElabObjectType, Callable[[], Iterable[ElabApiObjInterface]]]
        ] = [
            (ElabObjectType.EXPERIMENT, api.experiments.iter),
            (ElabObjectType.ITEM, api.items.iter),
            (ElabObjectType.ITEMS_TYPE, api.items_types.iter_full),
            (ElabObjectType.EXPERIMENTS_TEMPLATE, api.experiments_templates.iter),
        ]

        for obj_type, getter in fetch_jobs:
            logger.info(f"Pulling {obj_type.value} info...")
            objs = getter()
            for obj in objs:
                yield TypedObj(type=obj_type, obj=obj)

    @classmethod
    def _classify_objects(
        cls,
        api_objs: Iterable[ElabApiObjInterface],
    ) -> Iterator[TypedObj]:
        for api_obj in api_objs:
            obj_type = CLASS_TO_OBJTYPE.get(type(api_obj))
            if obj_type is None:
                logger.warning("Object %s could not be identified." % api_obj)
                continue
            yield TypedObj(type=obj_type, obj=api_obj)

    @classmethod
    def _enrich_objs(
        cls,
        objects: Iterable[TypedObj],
        metadata_parser: Callable[[str | None], MetadataModel],
        skip_untracked: bool,
    ) -> Iterator[EnrichedObj]:
        for typed_obj in objects:
            obj = typed_obj.obj

            if not obj.metadata and skip_untracked:
                continue

            enriched_obj = cls._process_obj(
                typed_obj=typed_obj,
                metadata_parser=metadata_parser,
            )

            if enriched_obj.metadata.elabftwcontrol is None and skip_untracked:
                continue

            yield enriched_obj

    @classmethod
    def _process_obj(
        cls,
        typed_obj: TypedObj,
        metadata_parser: Callable[[str | None], MetadataModel],
    ) -> EnrichedObj:
        return EnrichedObj.from_typed_obj(
            typed_obj=typed_obj,
            metadata_parser=metadata_parser,
        )


class StateManager:
    def __init__(
        self,
        state: State,
    ) -> None:
        self.state = state



import math
from datetime import date, datetime, time
from typing import Any, Callable, Optional

import pytest

from elabftwcontrol.core.models import (
    ConfigMetadata,
    FieldTypeEnum,
    GroupModel,
    MetadataModel,
    SingleFieldModel,
)
from elabftwcontrol.testing_utils import assert_dicts_equal


class TestSingleFieldModel:
    @pytest.mark.parametrize(
        ("input", "expected"),
        (
            ({}, SingleFieldModel()),
            (
                {
                    "type": "number",
                    "value": "2.3",
                    "unit": "kg",
                    "group_id": "1",
                    "position": "3",
                },
                SingleFieldModel(
                    type=FieldTypeEnum.number,
                    value="2.3",
                    unit="kg",
                    group_id=1,
                    position=3,
                ),
            ),
            (
                {
                    "value": ["select 1", "select 2"],
                    "type": "select",
                    "options": ["select 1", "select 2", "select 3"],
                    "group_id": "1",
                    "allow_multi_values": True,
                    "position": "3",
                    "blank_value_on_duplicate": True,
                    "readonly": True,
                },
                SingleFieldModel(
                    value=["select 1", "select 2"],
                    type=FieldTypeEnum.select,
                    options=["select 1", "select 2", "select 3"],
                    group_id=1,
                    allow_multi_values=True,
                    position=3,
                    blank_value_on_duplicate=True,
                    readonly=True,
                ),
            ),
        ),
    )
    def test_creation(self, input: dict[str, Any], expected: SingleFieldModel) -> None:
        assert SingleFieldModel(**input) == expected

    @pytest.mark.parametrize(
        ("value", "field_type", "expected"),
        (
            ("1234", FieldTypeEnum.number, 1234.0),
            ("1234-01-23", FieldTypeEnum.date, datetime(1234, 1, 23)),
            (
                "1234-01-23T12:34",
                FieldTypeEnum.datetime_local,
                datetime(1234, 1, 23, 12, 34),
            ),
            ("12:34", FieldTypeEnum.time, datetime(1900, 1, 1, 12, 34)),
            ("12", FieldTypeEnum.items, 12),
            ("12", FieldTypeEnum.experiments, 12),
            ("x", FieldTypeEnum.date, None),
            ("x", FieldTypeEnum.datetime_local, None),
            ("x", FieldTypeEnum.time, None),
            ("x", FieldTypeEnum.items, None),
            ("x", FieldTypeEnum.experiments, None),
            ("x", None, "x"),
            ("1234", None, "1234"),
        ),
    )
    def test_get_field_value(
        self,
        value: str,
        field_type: FieldTypeEnum,
        expected: Any,
    ) -> None:
        if field_type is None:
            field = SingleFieldModel(value=value)
        else:
            field = SingleFieldModel(value=value, type=field_type)
        assert field.parsed_value == expected

    @pytest.mark.parametrize(
        ("value", "expected", "test"),
        (
            ("1234", 1234.0, lambda result, expected: abs(result - expected) < 1e-6),
            ("x", float("nan"), lambda result, expected: math.isnan(result)),
        ),
    )
    def test_parse_number(
        self,
        value: str,
        expected: float,
        test: Callable[[float, float], bool],
    ) -> None:
        assert test(SingleFieldModel.parse_number(value), expected)

    @pytest.mark.parametrize(
        ("value", "expected"),
        (
            ("2024-05-20", datetime(2024, 5, 20)),
            ("x", None),
        ),
    )
    def test_parse_date(self, value: str, expected: Optional[date]) -> None:
        assert SingleFieldModel.parse_date(value) == expected

    @pytest.mark.parametrize(
        ("value", "expected"),
        (
            ("2024-05-20T19:05", datetime(2024, 5, 20, 19, 5)),
            ("x", None),
        ),
    )
    def test_parse_datetime(self, value: str, expected: Optional[datetime]) -> None:
        assert SingleFieldModel.parse_datetime(value) == expected

    @pytest.mark.parametrize(
        ("value", "expected"),
        (
            ("19:05", datetime(1900, 1, 1, 19, 5)),
            ("x", None),
        ),
    )
    def test_parse_time(self, value: str, expected: Optional[time]) -> None:
        assert SingleFieldModel.parse_time(value) == expected

    @pytest.mark.parametrize(
        ("value", "expected"),
        (
            ("19", 19),
            ("x", None),
        ),
    )
    def test_parse_item_link(self, value: str, expected: Optional[int]) -> None:
        assert SingleFieldModel.parse_item_link(value) == expected

    @pytest.mark.parametrize(
        ("value", "expected"),
        (
            ("19", 19),
            ("x", None),
        ),
    )
    def test_parse_experiment_link(self, value: str, expected: Optional[int]) -> None:
        assert SingleFieldModel.parse_experiment_link(value) == expected

    @pytest.mark.parametrize(
        ("value", "unit", "expected"),
        (
            ("1234", "kg", "kg"),
            (None, "kg", ""),
            ("1234", None, ""),
            (None, None, ""),
        ),
    )
    def test_get_field_unit(
        self,
        value: Any,
        unit: Optional[str],
        expected: str,
    ) -> None:
        field = SingleFieldModel(value=value, unit=unit)
        assert field.corrected_unit == expected

    @pytest.mark.parametrize(
        ("value", "unit", "expected"),
        (
            ("1234", "kg", "1234 kg"),
            (None, "kg", ""),
            ("1234", None, "1234"),
            (None, None, ""),
        ),
    )
    def test_get_field_combined(
        self,
        value: Any,
        unit: Optional[str],
        expected: str,
    ) -> None:
        field = SingleFieldModel(value=value, unit=unit)
        assert field.value_and_unit == expected


class TestMetadataModel:
    def test_creation(self) -> None:
        dummy_extra_data: dict[str, Any] = {
            "elabftw": {
                "extra_fields_groups": [
                    {"id": 1, "name": "g1"},
                    {"id": 2, "name": "g2"},
                    {"id": 3, "name": "g3"},
                ],
            },
            "extra_fields": {
                "f1": {
                    "type": "number",
                    "value": "2.3",
                    "unit": "kg",
                    "group_id": "1",
                    "position": "3",
                },
                "k3": {
                    "type": "date",
                    "value": "2023-09-12",
                    "group_id": "1",
                    "position": "2",
                },
            },
        }
        expected = MetadataModel(
            elabftw=ConfigMetadata(
                extra_fields_groups=[
                    GroupModel(id=1, name="g1"),
                    GroupModel(id=2, name="g2"),
                    GroupModel(id=3, name="g3"),
                ]
            ),
            extra_fields={
                "f1": SingleFieldModel(
                    type=FieldTypeEnum.number,
                    value="2.3",
                    unit="kg",
                    group_id=1,
                    position=3,
                ),
                "k3": SingleFieldModel(
                    type=FieldTypeEnum.date,
                    value="2023-09-12",
                    group_id=1,
                    position=2,
                ),
            },
        )
        model = MetadataModel(**dummy_extra_data)
        assert expected == model

    def test_create_empty(self) -> None:
        expected = MetadataModel()
        model = MetadataModel(**{})
        assert expected == model

    @pytest.mark.parametrize(
        ("wrong_input",),
        (
            (
                {
                    "extra_fields": {
                        "f2": {
                            "type": "non_existing_type",
                        },
                    },
                },
            ),
            (
                {
                    "extra_fields": {
                        "f2": {
                            "group_id": "not_integer",
                        },
                    },
                },
            ),
            (
                {
                    "extra_fields": {
                        "f2": {
                            "position": "not_integer",
                        },
                    },
                },
            ),
        ),
    )
    def test_metadatamodel_failure(self, wrong_input: dict[str, Any]) -> None:
        with pytest.raises(Exception):
            _ = MetadataModel(**wrong_input)

    def test_ordered_fieldnames(self) -> None:
        wrong_field: dict[str, Any] = {}
        model = MetadataModel(
            extra_fields={
                "f1": SingleFieldModel(position=3),
                "f2": SingleFieldModel(position=1),
                "f3": SingleFieldModel(position=2),
                "f4": SingleFieldModel(position=None),
                "f5": SingleFieldModel(**wrong_field),
            },
        )
        expected = ["f4", "f5", "f2", "f3", "f1"]
        assert model.ordered_fieldnames == expected

    def test_fieldnames(self) -> None:
        wrong_field: dict[str, Any] = {}
        model = MetadataModel(
            extra_fields={
                "f1": SingleFieldModel(position=3),
                "f2": SingleFieldModel(position=1),
                "f3": SingleFieldModel(position=2),
                "f4": SingleFieldModel(position=None),
                "f5": SingleFieldModel(**wrong_field),
            },
        )
        expected = ["f1", "f2", "f3", "f4", "f5"]
        assert model.fieldnames == expected

    def test_field_values(self) -> None:
        wrong_field: dict[str, Any] = {
            "value": "123",
            "type": "number",
        }
        wrong_field_2: dict[str, Any] = {
            "value": "not a number",
            "type": "number",
            "position": "-1",
        }
        model = MetadataModel(
            extra_fields={
                "f1": SingleFieldModel(
                    value="1993-06-18",
                    type=FieldTypeEnum.date,
                    position=3,
                ),
                "f2": SingleFieldModel(value="1", position=1),
                "f3": SingleFieldModel(
                    value="234",
                    type=FieldTypeEnum.items,
                    position=2,
                ),
                "f4": SingleFieldModel(value="value4", position=None),
                "f5": SingleFieldModel(**wrong_field),
                "f6": SingleFieldModel(**wrong_field_2),
            },
        )
        expected = {
            "f4": "value4",
            "f5": 123,
            "f6": float("nan"),
            "f2": "1",
            "f3": 234,
            "f1": datetime(1993, 6, 18),
        }
        assert_dicts_equal(model.field_values, expected)

    def test_field_units(self) -> None:
        wrong_field: dict[str, Any] = {
            "value": "123",
            "unit": "unit_5",
            "type": "number",
        }
        wrong_field_2: dict[str, Any] = {
            "value": "not a number",
            "unit": None,
            "type": "number",
            "position": "-1",
        }
        model = MetadataModel(
            extra_fields={
                "f1": SingleFieldModel(
                    value="1993-06-18",
                    unit="unit_1",
                    type=FieldTypeEnum.date,
                    position=3,
                ),
                "f2": SingleFieldModel(
                    value="1",
                    position=1,
                ),
                "f3": SingleFieldModel(
                    value="234 - something - something",
                    type=FieldTypeEnum.items,
                    position=2,
                ),
                "f4": SingleFieldModel(
                    value="value4",
                    unit="unit_4",
                    position=None,
                ),
                "f5": SingleFieldModel(**wrong_field),
                "f6": SingleFieldModel(**wrong_field_2),
            },
        )
        expected = {
            "f4": "unit_4",
            "f5": "unit_5",
            "f6": "",
            "f2": "",
            "f3": "",
            "f1": "unit_1",
        }
        assert_dicts_equal(model.field_units, expected)

    def test_field_value_and_units(self) -> None:
        wrong_field: dict[str, Any] = {
            "value": "123",
            "unit": "unit_5",
            "type": "number",
        }
        wrong_field_2: dict[str, Any] = {
            "value": "not a number",
            "unit": None,
            "type": "number",
            "position": "-1",
        }
        model = MetadataModel(
            extra_fields={
                "f1": SingleFieldModel(
                    value="1993-06-18",
                    unit="unit_1",
                    type=FieldTypeEnum.date,
                    position=3,
                ),
                "f2": SingleFieldModel(
                    value="1",
                    position=1,
                ),
                "f3": SingleFieldModel(
                    value="234 - something - something",
                    type=FieldTypeEnum.items,
                    position=2,
                ),
                "f4": SingleFieldModel(
                    value="value4",
                    unit="unit_4",
                    position=None,
                ),
                "f5": SingleFieldModel(**wrong_field),
                "f6": SingleFieldModel(**wrong_field_2),
            },
        )
        expected = {
            "f4": "value4 unit_4",
            "f5": "123 unit_5",
            "f6": "not a number",
            "f2": "1",
            "f3": "234 - something - something",
            "f1": "1993-06-18 unit_1",
        }
        assert model.field_values_and_units == expected

    def test_get_extra_field_names_in_order(self) -> None:
        metadata: dict[str, Any] = {
            "extra_fields": {
                "field 3": {"position": "10"},
                "field 1": {"position": "8"},
                "field 2": {},
            }
        }
        parsed_metadata = MetadataModel(**metadata)
        expected = ["field 2", "field 1", "field 3"]
        assert parsed_metadata.ordered_fieldnames == expected



from typing import Any, Collection, Mapping, NamedTuple

from typing_extensions import Self


class Change(NamedTuple):
    old: Any
    new: Any

    def __str__(self) -> str:
        return f"{self.old} -> {self.new}"


def _dict_to_str(
    data: Mapping[str, Any],
    marker: str,
    indent: int,
) -> str:
    result = ""
    for k, v in data.items():
        result += " " * indent + f"{marker} {k}: {v}\n"
    return result[:-1]


def _sequence_to_str(
    data: Collection[str],
    marker: str,
    indent: int,
) -> str:
    result = ""
    for k in data:
        result += " " * indent + f"{marker} {k}\n"
    return result[:-1]


_RED = "\033[31m"
_GREEN = "\033[32m"
_YELLOW = "\033[33m"
_RESET = "\033[0m"


class FieldsToAdd(NamedTuple):
    data: Mapping[str, Any]

    def __bool__(self) -> bool:
        if self.data:
            return True
        else:
            return False

    def to_str(self, indent: int = 0) -> str:
        return _dict_to_str(data=self.data, marker=f"{_GREEN}+{_RESET}", indent=indent)


class FieldsToChange(NamedTuple):
    data: Mapping[str, Change]

    def __bool__(self) -> bool:
        if self.data:
            return True
        else:
            return False

    def to_str(self, indent: int = 0) -> str:
        return _dict_to_str(data=self.data, marker=f"{_YELLOW}~{_RESET}", indent=indent)


class FieldsToDelete(NamedTuple):
    data: Collection[str]

    def __bool__(self) -> bool:
        if self.data:
            return True
        else:
            return False

    def to_str(self, indent: int = 0) -> str:
        return _sequence_to_str(
            data=self.data, marker=f"{_RED}-{_RESET}", indent=indent
        )


class DictComparisonResult(NamedTuple):
    to_add: FieldsToAdd
    to_change: FieldsToChange
    to_delete: FieldsToDelete

    def __bool__(self) -> bool:
        if self.to_add or self.to_change or self.to_delete:
            return True
        else:
            return False

    def to_str(self, indent: int = 0) -> str:
        results = []
        for collection in (self.to_add, self.to_change, self.to_delete):
            if collection:
                results.append(collection.to_str(indent))
        return "\n".join(results)

    @classmethod
    def from_dicts(
        cls,
        old: dict[str, Any],
        new: dict[str, Any],
    ) -> Self:
        keys_old = set(old.keys())
        keys_new = set(new.keys())

        keys_to_add = keys_new - keys_old
        to_add = {k: new[k] for k in keys_to_add}

        to_delete = keys_old - keys_new

        to_check = keys_new.intersection(keys_old)
        to_change: dict[str, Change] = {}
        for k in to_check:
            if old[k] == new[k]:
                continue
            to_change[k] = Change(old=old[k], new=new[k])

        return cls(
            to_add=FieldsToAdd(to_add),
            to_change=FieldsToChange(to_change),
            to_delete=FieldsToDelete(to_delete),
        )


class MetadataFieldsToAdd(NamedTuple):
    data: Mapping[str, FieldsToAdd]

    def __bool__(self) -> bool:
        if self.data:
            return True
        else:
            return False

    def to_str(self, indent: int = 0) -> str:
        result = ""
        for field_name, field_info in self.data.items():
            result += (
                " " * indent
                + f"""\
{_GREEN}+{_RESET} {field_name}:
{field_info.to_str(indent + 4)}
"""
            )
        return result


class MetadataFieldsToChange(NamedTuple):
    data: Mapping[str, DictComparisonResult]

    def __bool__(self) -> bool:
        if self.data:
            return True
        else:
            return False

    def to_str(self, indent: int = 0) -> str:
        result = ""
        for field_name, field_info in self.data.items():
            result += (
                " " * indent
                + f"""\
{_YELLOW}~{_RESET} {field_name}:
{field_info.to_str(indent + 4)}
"""
            )
        return result


class MetadataFieldsToDelete(NamedTuple):
    data: Collection[str]

    def __bool__(self) -> bool:
        if self.data:
            return True
        else:
            return False

    def to_str(self, indent: int = 0) -> str:
        return _sequence_to_str(
            data=self.data, marker=f"{_RED}-{_RESET}", indent=indent
        )


class MetadataComparisonResult(NamedTuple):
    to_add: MetadataFieldsToAdd
    to_change: MetadataFieldsToChange
    to_delete: MetadataFieldsToDelete

    def __bool__(self) -> bool:
        if self.to_add or self.to_change or self.to_delete:
            return True
        else:
            return False

    def to_str(self, indent: int = 0) -> str:
        result = ""
        for collection in (self.to_add, self.to_change, self.to_delete):
            if collection:
                result += collection.to_str(indent)
        return result

    @classmethod
    def from_dicts(
        cls,
        old: dict[str, dict[str, Any]],
        new: dict[str, dict[str, Any]],
    ) -> Self:
        keys_old = set(old.keys())
        keys_new = set(new.keys())

        keys_to_add = keys_new - keys_old
        to_add = {k: FieldsToAdd(new[k]) for k in keys_to_add}

        to_delete = keys_old - keys_new

        to_check = keys_new.intersection(keys_old)
        to_change: dict[str, DictComparisonResult] = {}
        for k in to_check:
            diff = DictComparisonResult.from_dicts(old[k], new[k])
            if diff:
                to_change[k] = diff

        return cls(
            to_add=MetadataFieldsToAdd(to_add),
            to_change=MetadataFieldsToChange(to_change),
            to_delete=MetadataFieldsToDelete(to_delete),
        )


class Diff(NamedTuple):
    main_fields: DictComparisonResult
    metadata_fields: MetadataComparisonResult

    def __bool__(self) -> bool:
        if self.main_fields or self.metadata_fields:
            return True
        else:
            return False

    def to_str(self, indent: int = 0) -> str:
        result = ""
        shift = " " * indent
        if self.main_fields:
            result += f"""\
{shift}Main fields
{shift}-----------
{self.main_fields.to_str(indent)}

"""
        if self.metadata_fields:
            result += f"""\
{shift}Metadata fields
{shift}---------------
{self.metadata_fields.to_str(indent)}"""
        return result.strip("\n")

    @classmethod
    def new(
        cls,
        old: dict[str, Any],
        new: dict[str, Any],
        old_metadata_fields: dict[str, dict[str, Any]],
        new_metadata_fields: dict[str, dict[str, Any]],
    ) -> Self:
        main_fields = DictComparisonResult.from_dicts(old=old, new=new)
        metadata_fields = MetadataComparisonResult.from_dicts(
            old=old_metadata_fields,
            new=new_metadata_fields,
        )
        return cls(
            main_fields=main_fields,
            metadata_fields=metadata_fields,
        )



import logging
import warnings
from enum import Enum
from pathlib import Path
from typing import Optional

import typer
from typing_extensions import Annotated

from elabftwcontrol._logging import logger
from elabftwcontrol.client import ElabftwApi
from elabftwcontrol.configure import (
    create_or_append_configuration_file,
    delete_configuration_file,
    list_config_profiles,
)
from elabftwcontrol.defaults import DEFAULT_CONFIG_FILE
from elabftwcontrol.download import (
    IngestConfiguration,
    IngestJob,
    ObjectTypes,
    OutputFormats,
    TableCellContentType,
    TableShapes,
)
from elabftwcontrol.upload.api import ApplyJob, ApplyJobConfig, ApplyJobType

CMD = "elabftwctl"

app = typer.Typer(
    name=CMD,
    no_args_is_help=True,
    help="CLI utility and library for interacting with eLabFTW",
)


class ConfigActions(str, Enum):
    set = "set"
    list = "list"
    delete = "delete"


@app.command(
    help="Interact with the configuration for connecting to eLabFTW",
    no_args_is_help=True,
)
def config(
    action: Annotated[
        ConfigActions,
        typer.Argument(help="Action to perform on the config."),
    ],
    filepath: Annotated[
        Optional[Path],
        typer.Option(
            help="Path where config details exist or will be stored.",
        ),
    ] = None,
    profile: Annotated[
        Optional[str],
        typer.Option(
            help="Name of the profile. All by default or 'default' on set.",
        ),
    ] = None,
    show_keys: Annotated[
        bool,
        typer.Option(
            "--show_keys",
            help="Reveal the API keys in the terminal.",
        ),
    ] = False,
) -> None:
    if filepath is None:
        filepath = DEFAULT_CONFIG_FILE
    match action.value:
        case ConfigActions.set:
            create_or_append_configuration_file(
                filepath=filepath,
                profile=profile,
                show_keys=show_keys,
            )
        case ConfigActions.list:
            list_config_profiles(
                filepath=filepath,
                profile=profile,
                show_keys=show_keys,
            )
        case ConfigActions.delete:
            delete_configuration_file(
                filepath=filepath,
                profile=profile,
                show_keys=show_keys,
            )
        case _:
            print("Action not recognized")
            raise typer.Abort()


def _get_api(profile: Optional[str]) -> ElabftwApi:
    if profile:
        try:
            api = ElabftwApi.from_config_file(profile=profile)
        except Exception:
            raise ValueError(
                f"No profile with name {profile} was found. "
                "Please configure it using 'elabftwctl config set'."
            )
    else:
        try:
            api = ElabftwApi.new()
        except ValueError:
            try:
                api = ElabftwApi.from_config_file(profile="default")
            except Exception:
                raise ValueError(
                    "No default profile was found. "
                    "Please configure it using 'elabftwctl config set'."
                )
    return api


def _set_logger_verbosity(verbosity: int) -> None:
    if verbosity == 0:
        logger.setLevel(logging.WARN)
    elif verbosity == 1:
        logger.setLevel(logging.INFO)
    else:
        logger.setLevel(logging.DEBUG)


@app.command(
    help="Download or ingest data from eLabFTW in various formats",
    no_args_is_help=True,
)
def get(
    object_type: Annotated[
        ObjectTypes,
        typer.Argument(help="Which type of resource to ingest."),
    ],
    format: Annotated[
        OutputFormats,
        typer.Option(
            help="Data format in which the output is represented.",
        ),
    ] = OutputFormats.JSON,
    id: Annotated[
        list[int],
        typer.Option(
            help="Limit the output to these ids.",
        ),
    ] = [],
    output: Annotated[
        Optional[str],
        typer.Option(
            help=(
                "Path where the results should be directed. Can be a path on the filesystem "
                "or a location on AWS S3 if prefixed by s3://. If not provided defaults to stdout."
            ),
        ),
    ] = None,
    category: Annotated[
        list[str],
        typer.Option(
            help="Limit the output to these categories or templates. Provide titles as input.",
        ),
    ] = [],
    indent: Annotated[
        bool,
        typer.Option(
            help="Whether to indent JSON results.",
        ),
    ] = False,
    expand_metadata: Annotated[
        bool,
        typer.Option(
            help=(
                "In tabular formats, whether to parse and expand the metadata column."
            ),
        ),
    ] = False,
    table_shape: Annotated[
        TableShapes,
        typer.Option(
            help=(
                "In tabular formats, whether or to expand the metadata fields "
                "into columns or stack the information in rows."
            ),
        ),
    ] = TableShapes.WIDE,
    use_template_metadata_schema: Annotated[
        bool,
        typer.Option(
            help=(
                "In wide tabular formats, whether to limit the columns to "
                "the fields from the corresponding templates."
            ),
        ),
    ] = False,
    cell_content: Annotated[
        TableCellContentType,
        typer.Option(
            help=(
                "In wide tabular formats, what field information to use to populate the "
                "metadata columns."
            ),
        ),
    ] = TableCellContentType.VALUE,
    sanitize_column_names: Annotated[
        bool,
        typer.Option(
            help="Convert column names to a limited character set.",
        ),
    ] = False,
    glue_table: Annotated[
        Optional[str],
        typer.Option(
            help=(
                "If you write single table data to S3, you can register it in glue. "
                "Use the format <DB NAME>:<TABLE NAME>."
            ),
        ),
    ] = None,
    profile: Annotated[
        Optional[str],
        typer.Option(
            help=(
                f"Name of the eLabFTW connection profile to use, see `{CMD} config`. "
                "If not set, 'default' is used."
            ),
        ),
    ] = None,
    verbosity: Annotated[
        int,
        typer.Option(
            "--verbose",
            "-v",
            count=True,
            help="Increase verbosity level.",
        ),
    ] = 0,
) -> None:
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        api = _get_api(profile)
        _set_logger_verbosity(verbosity)

        try:
            config = IngestConfiguration(
                object_type=object_type,
                ids=id,
                format=format,
                output=output,
                categories=category,
                indent=indent,
                expand_metadata=expand_metadata,
                table_shape=table_shape,
                use_template_metadata_schema=use_template_metadata_schema,
                cell_content=cell_content,
                sanitize_column_names=sanitize_column_names,
                glue_table=glue_table,
            )
            logger.debug("Inputs:\n %s" % config.model_dump_json(indent=2))
            IngestJob(api=api, config=config)()
        except Exception as e:
            logger.critical("%s" % e)


@app.command(
    help="Change the state in eLabFTW through state defined in YAML manifest files.",
    no_args_is_help=True,
)
def apply(
    folder: Annotated[
        Path,
        typer.Argument(help="Folder in which the manifest files live."),
    ],
    n_retries: Annotated[
        int,
        typer.Option(
            help="Maximum number of times applies should be retried after failing."
        ),
    ] = 3,
    no_prompt: Annotated[
        bool,
        typer.Option(help=("Do not ask for confirmation.")),
    ] = False,
    version: Annotated[
        str | None,
        typer.Option(
            help=("Version identifier that should be used to annotate objects.")
        ),
    ] = None,
    dry_run: Annotated[
        bool,
        typer.Option(help=("Do not actually apply, just check the changes.")),
    ] = False,
    destroy: Annotated[
        bool,
        typer.Option(
            "--destroy",
            "-d",
            help="Instead of applying the manifests, destroy all the matching state.",
        ),
    ] = False,
    # state_file: Annotated[
    #     Path,
    #     typer.Option(help="Instead of downloading state from eLabFTW it is passed in a file."),
    # ],
    # ignore_state: Annotated[
    #     bool,
    #     typer.Option(
    #         help=(
    #             "Do not check the current state in eLabFTW and assume it is empty. "
    #             "This can give rise to duplicates!"
    #         ),
    #     ),
    # ] = False,
    profile: Annotated[
        Optional[str],
        typer.Option(
            help=(
                f"Name of the eLabFTW connection profile to use, see `{CMD} config`. "
                "If not set, 'default' is used."
            ),
        ),
    ] = None,
    verbosity: Annotated[
        int,
        typer.Option(
            "--verbose",
            "-v",
            count=True,
            help="Increase verbosity level.",
        ),
    ] = 0,
) -> None:
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        api = _get_api(profile)
        _set_logger_verbosity(verbosity)

        if destroy:
            job_type = ApplyJobType.DESTROY
        else:
            job_type = ApplyJobType.CREATE

        try:
            config = ApplyJobConfig(
                jobtype=job_type,
                manifest_folder=folder,
                ignore_state=False,
                state_file=None,
                version=version,
                n_retries=n_retries,
                no_prompt=no_prompt,
                dry_run=dry_run,
            )
            ApplyJob(api, config)()
        except Exception as e:
            logger.critical("%s" % e)


def main() -> None:
    app()


if __name__ == "__main__":
    main()



import json
from typing import Any

from elabftwcontrol._logging import logger
from elabftwcontrol.core.models import MetadataModel


class MetadataParser:
    """Parses metadata string from an API object"""

    def __call__(self, metadata: str | None) -> MetadataModel:
        return MetadataModel(**self._safe_parse_metadata(metadata))

    @classmethod
    def _safe_parse_metadata(
        cls,
        obj_metadata: str | None,
    ) -> dict[str, Any]:
        if obj_metadata is None:
            return {}

        try:
            data = json.loads(obj_metadata)
        except (TypeError, json.JSONDecodeError) as e:
            logger.warn(f"Metadata obj object could not be parsed as valid JSON:\n{e}")
            return {}

        return data


class TagParser:
    """Parses tags from an API object"""

    def __call__(self, data: str | None) -> list[str]:
        if data is None:
            return []
        try:
            return data.split("|")
        except Exception:
            logger.warn(f"Data `{data}` could not be parsed as tags")
        return []


class TagIdParser:
    """Parses tag ids from an API object"""

    def __call__(self, data: str | None) -> list[int]:
        if data is None:
            return []
        try:
            return [int(tag_id.strip()) for tag_id in data.split(",")]
        except Exception:
            logger.warn(f"Data `{data}` could not be parsed as tag ids")
        return []



from __future__ import annotations

import json
from functools import cached_property
from itertools import chain
from typing import (
    Any,
    Callable,
    Collection,
    Iterable,
    Iterator,
    NamedTuple,
    Optional,
    Sequence,
)

import pandas as pd
from pydantic import BaseModel, model_validator

from elabftwcontrol._logging import logger
from elabftwcontrol.client import ElabftwApi, Experiment, Item
from elabftwcontrol.core.interfaces import Category, Dictable
from elabftwcontrol.core.metadata import ParsedMetadataToPandasDtype
from elabftwcontrol.core.parsers import MetadataParser
from elabftwcontrol.core.transformers import (
    JSONTransformer,
    MultiPandasDataFrameTransformer,
    ObjectTypes,
    PandasDataFrameMetadataTransformer,
    PandasDataFrameTransformer,
    SplitDataFrame,
    TableCellContentType,
    TableShapes,
)
from elabftwcontrol.download.output import (
    ExcelWriter,
    LineWriter,
    OutputFormats,
    PandasWriter,
)
from elabftwcontrol.global_config import GlobalConfig


def get_item_types_as_df(
    client: Optional[ElabftwApi] = None,
    ids: Optional[int | Sequence[int]] = None,
) -> pd.DataFrame:
    """Get a table representation of the item types / categories

    Parameters
    ----------
    client
        A custom ElabftwApi object.
    ids
        Limit the returned records to these ids. By default all are included.

    Returns
    -------
    table
        The items type data in pandas table format. Fields are mapped to columns.
    """
    if isinstance(ids, int):
        ids = [ids]

    client = _client_or_default(client)
    item_types = client.items_types.iter_full(category_ids=ids)
    return _convert_to_table(
        objects=item_types,
        obj_type=ObjectTypes.ITEMS_TYPE,
        expand_metadata=False,
    )


def get_experiment_templates_as_df(
    client: Optional[ElabftwApi] = None,
    ids: Optional[int | Sequence[int]] = None,
) -> pd.DataFrame:
    """Get a table representation of the experiment templates

    Parameters
    ----------
    client
        A custom ElabftwApi object.
    ids
        Limit the returned records to these ids. By default all are included.

    Returns
    -------
    table
        The items type data in pandas table format. Fields are mapped to columns.
    """
    if isinstance(ids, int):
        ids = [ids]

    client = _client_or_default(client)
    experiment_templates = client.experiments_templates.iter(ids=ids)
    return _convert_to_table(
        objects=experiment_templates,
        obj_type=ObjectTypes.EXPERIMENTS_TEMPLATE,
        expand_metadata=False,
    )


def get_experiments_as_df(
    client: Optional[ElabftwApi] = None,
    ids: Optional[int | Collection[int]] = None,
    template_names: Optional[str | Sequence[str]] = None,
    expand_metadata: bool = False,
    table_shape: TableShapes = TableShapes.WIDE,
    cell_content: TableCellContentType = TableCellContentType.VALUE,
    sanitize_column_names: bool = False,
) -> pd.DataFrame:
    """Get a table representation of the experiments

    Parameters
    ----------
    client
        A custom ElabftwApi object.
    ids
        Limit the returned records to these ids. By default all are included.
        Order may not be conserved.
    template_names
        Experiments derived from these templates should be included. By default
        all experiments are included.
    expand_metadata
        Processes the table to expand the metadata information to columns and
        rows. This also makes a selection of the original columns and prefixes them
        with an underscore.
    table_shape
        Only relevant if `expand_metadata=True`. Options are "wide" and "long".
        "long" stacks all metadata information into separate rows, meaning that each
        experiment with n metadata fields is expanded to n rows. All field values stay
        represented as strings. "wide" creates new columns for metadata information
        and conserves one row per experiment. Data types are automatically inferred
        from the field type.
    cell_content
        Only relevant if "expand_metadata=True` and `table_shape="wide"`. By default
        only the field values are stored in the metadata field columns. This can become
        ambiguous if the field also had units. Setting this argument to `"combined"` will
        retain both value and unit as a string. It is also possible to select only `"unit`.
    sanitize_column_names
        Only relevant if "expand_metadata=True` and `table_shape="wide"`. The metadata
        column names correspond to field names. Setting this option to `True` will
        convert those names to a limited character set.

    Returns
    -------
    table
        The experiment data in pandas table format. Fields are mapped to columns.

    Notes
    -----
    * the `template_names` parameter works by substring search. It is assumed that
      the title of the template is still part of the experiment title. If this is
      not the case, there is no reliable way to determine which template an experiment
      originated from.
    """
    client = _client_or_default(client)

    if isinstance(template_names, str):
        template_names = [template_names]

    if isinstance(ids, int):
        ids = [ids]

    experiments = IngestJob.get_experiments(
        client,
        ids=ids,
        template_names=template_names,
    )
    return _convert_to_table(
        objects=experiments,
        obj_type=ObjectTypes.EXPERIMENT,
        expand_metadata=expand_metadata,
        table_shape=table_shape,
        cell_content=cell_content,
        sanitize_column_names=sanitize_column_names,
    )


def get_items_as_df(
    client: Optional[ElabftwApi] = None,
    ids: Optional[int | Collection[int]] = None,
    category_ids: Optional[int | Sequence[int]] = None,
    expand_metadata: bool = False,
    table_shape: TableShapes = TableShapes.WIDE,
    cell_content: TableCellContentType = TableCellContentType.VALUE,
    sanitize_column_names: bool = False,
) -> pd.DataFrame:
    """Get a table representation of the items

    Parameters
    ----------
    client
        A custom ElabftwApi object.
    ids
        Limit the returned records to these ids. By default all are included.
        Order may not be conserved.
    category_ids
        Items derived from only these categories should be included. By default
        all items are included.
    expand_metadata
        Processes the table to expand the metadata information to columns and
        rows. This also makes a selection of the original columns and prefixes them
        with an underscore.
    table_shape
        Only relevant if `expand_metadata=True`. Options are "wide" and "long".
        "long" stacks all metadata information into separate rows, meaning that each
        item with n metadata fields is expanded to n rows. All field values stay
        represented as strings. "wide" creates new columns for metadata information
        and conserves one row per item. Data types are automatically inferred
        from the field type.
    cell_content
        Only relevant if "expand_metadata=True` and `table_shape="wide"`. By default
        only the field values are stored in the metadata field columns. This can become
        ambiguous if the field also had units. Setting this argument to `"combined"` will
        retain both value and unit as a string. It is also possible to select only `"unit`.
    sanitize_column_names
        Only relevant if "expand_metadata=True` and `table_shape="wide"`. The metadata
        column names correspond to field names. Setting this option to `True` will
        convert those names to a limited character set.

    Returns
    -------
    table
        The experiment data in pandas table format. Fields are mapped to columns.
    """
    client = _client_or_default(client)

    if isinstance(category_ids, int):
        category_ids = [category_ids]

    if isinstance(ids, int):
        ids = [ids]

    items = IngestJob.get_items(
        client,
        ids=ids,
        category_ids=category_ids,
    )
    return _convert_to_table(
        objects=items,
        obj_type=ObjectTypes.ITEM,
        expand_metadata=expand_metadata,
        table_shape=table_shape,
        cell_content=cell_content,
        sanitize_column_names=sanitize_column_names,
    )


def _client_or_default(client: Optional[ElabftwApi]) -> ElabftwApi:
    if client is None:
        client = GlobalConfig.get_client()
    return client


def _convert_to_table(
    objects: Iterable[Dictable],
    obj_type: ObjectTypes,
    expand_metadata: bool = True,
    table_shape: TableShapes = TableShapes.WIDE,
    cell_content: TableCellContentType = TableCellContentType.VALUE,
    sanitize_column_names: bool = False,
) -> pd.DataFrame:
    raw_table = PandasDataFrameTransformer.new()(objects)
    if expand_metadata:
        if table_shape == TableShapes.WIDE:
            transformer = PandasDataFrameMetadataTransformer.for_wide_table(
                object_type=obj_type,
                cell_content=cell_content,
                metadata_schema=None,
                sanitize_column_names=sanitize_column_names,
            )
        else:
            transformer = PandasDataFrameMetadataTransformer.for_long_table(
                object_type=obj_type,
            )

        return transformer(raw_table)
    else:
        return raw_table


class IngestConfiguration(BaseModel):
    """Parameters for downloading and storing data from eLabFTW

    Parameters
    ----------
    object_type
        Type of object to get from the server.
        Options: `"item"`, `"experiment"`, `"items_type"`, `"experiments_template"`
    ids
        Limit the output to records with these ids. Order may not be conserved.
    format
        Output formats. Available are `"json"`, `"csv"`, `"parquet"`, and `"excel"`.
        In the case of `"excel"`, records are split into different sheets by category.
    output
        Path to a file where the data should be stored. Defaults to stdout. Typically a path
        to a local file. If you have the AWS dependencies installed and you select
        csv or parquet, you can write directly to s3 using a path prefixed by
        s3://<bucket>/<prefix>.
    categories
        The items types or templates of the items or experiments to download. By default all
        objects are downloaded. For single table formats (csv, parquet) in wide format, it is
        recommended to select individual categories. For experiments, the template title
        must appear in the template of the experiment.
    indent
        Only applied when `format="json"`. Whether to indent json data that is returned.
    expand_metadata
        Only relevant to tabular output formats. If `True`, the data in the metadata column
        will be expanded to multiple columns and/or rows.
    table_shape
        Only relevant to tabular output formats. Defines whether metadata information is
        expanded to individual columns (wide) or stacked in rows (long).
    cell_content
        For wide tabular formats, the information that is put in columns corresponding to
        metadata fields. By default only the field value is conserved in the columns.
        This may be ambiguous for fields containing units. If `combined` is selected then
        field and unit are combined as a string. It is also possible to select only the unit.
    use_template_metadata_schema
        Only relevant to wide tabular formats. If true, the metadata columns are forced
        correspond to the fields of the item category or experiment template.
    sanitize_column_names
        Only relevant to wide tabular formats. Whether to keep the field names as column
        names or sanitize them to a simple character set that is compliant with column name
        rules in AWS Glue tables.
    glue_table
        Only relevant to tabular formats. If you write data directly to S3 you can register
        the table as a glue table. Input must be in the format: `<glue db>:<glue table>`.
    """

    object_type: ObjectTypes = ObjectTypes.ITEM
    ids: list[int] = []
    format: OutputFormats = OutputFormats.JSON
    output: Optional[str] = None
    categories: list[str] = []
    indent: bool = False
    expand_metadata: bool = False
    table_shape: TableShapes = TableShapes.WIDE
    use_template_metadata_schema: bool = False
    cell_content: TableCellContentType = TableCellContentType.VALUE
    sanitize_column_names: bool = False
    glue_table: Optional[str] = None

    @model_validator(mode="after")
    def check_consistency(self) -> IngestConfiguration:
        self._check_one_category_wide_single_table()
        return self

    def _check_one_category_wide_single_table(self) -> None:
        if (
            self.format in (OutputFormats.PARQUET, OutputFormats.CSV)
            and self.table_shape == TableShapes.WIDE
            and (not self.categories or len(self.categories) != 1)
            and self.expand_metadata
            and self.use_template_metadata_schema
        ):
            raise ValueError(
                "A single wide table format with metadata must specify one category"
            )


class IngestJob:
    def __init__(
        self,
        api: ElabftwApi,
        config: IngestConfiguration,
    ) -> None:
        self.api = api
        self.config = config

    def __call__(self) -> None:
        object_type = self.config.object_type
        logger.info("Starting ingestion for %s" % object_type)
        match object_type:
            case ObjectTypes.ITEM:
                self.ingest_items_data(api=self.api, config=self.config)
            case ObjectTypes.EXPERIMENT:
                self.ingest_experiments_data(api=self.api, config=self.config)
            case ObjectTypes.ITEMS_TYPE:
                self.ingest_items_types_data(api=self.api, config=self.config)
            case ObjectTypes.EXPERIMENTS_TEMPLATE:
                self.ingest_experiments_template_data(api=self.api, config=self.config)
            case _:
                raise NotImplementedError(
                    f"Ingestion for object type '{object_type}' not implemented"
                )
        logger.info("Completed ingestion")

    @classmethod
    def ingest_experiments_data(
        cls,
        api: ElabftwApi,
        config: IngestConfiguration,
    ) -> None:
        category_info = CategoryInfo(getter=api.experiments_templates.iter)
        experiments = cls.get_experiments(
            api,
            ids=config.ids,
            template_names=config.categories,
        )

        cls.transform_and_save_objects(
            objects=experiments,
            category_info=category_info,
            config=config,
        )

    @classmethod
    def ingest_experiments_template_data(
        cls,
        api: ElabftwApi,
        config: IngestConfiguration,
    ) -> None:
        def throw_error() -> Any:
            raise ValueError(
                "This operation requires information on categories, but "
                "experiment templates are categories themselves."
            )

        category_info = CategoryInfo(getter=throw_error)

        if config.categories:
            templates_info = CategoryInfo(
                getter=lambda: api.experiments_templates.iter(ids=config.ids),
            )
            template_ids = templates_info.get_ids(categories=config.categories)
            templates = [
                templates_info.categories_map[template_id]
                for template_id in template_ids
            ]
        else:
            templates = list(api.experiments_templates.iter(ids=config.ids))

        cls.transform_and_save_objects(
            objects=templates,
            category_info=category_info,
            config=config,
        )

    @classmethod
    def ingest_items_data(
        cls,
        api: ElabftwApi,
        config: IngestConfiguration,
    ) -> None:
        if not config.categories:
            included_item_ids = None
        else:
            included_item_ids = CategoryInfo(api.items_types.iter).get_ids(
                categories=config.categories
            )
            logger.debug("Item types categories specified: %s" % config.categories)
            logger.debug("Item types ids included: %s" % included_item_ids)

        items = cls.get_items(
            api,
            ids=config.ids,
            category_ids=included_item_ids,
        )

        category_info = CategoryInfo(api.items_types.iter_full)
        cls.transform_and_save_objects(
            objects=items,
            category_info=category_info,
            config=config,
        )

    @classmethod
    def ingest_items_types_data(
        cls,
        api: ElabftwApi,
        config: IngestConfiguration,
    ) -> None:
        def throw_error() -> Any:
            raise ValueError(
                "This operation requires information on categories, but "
                "items types are categories themselves."
            )

        category_info = CategoryInfo(getter=throw_error)

        items_types_info = CategoryInfo(lambda: api.items_types.iter(ids=config.ids))
        category_ids = items_types_info.get_ids(config.categories)
        item_types = api.items_types.iter_full(category_ids)

        cls.transform_and_save_objects(
            objects=item_types,
            category_info=category_info,
            config=config,
        )

    @classmethod
    def transform_and_save_objects(
        cls,
        objects: Iterable[Dictable],
        category_info: CategoryInfo,
        config: IngestConfiguration,
    ) -> None:
        if config.format == OutputFormats.JSON:
            logger.info("Ingesting as JSON")
            JSONConverter(config)(objects)
        else:
            if config.format == OutputFormats.EXCEL:
                logger.info(
                    "Attempting to process metadata, convert to multiple dataframes, "
                    "and save to Excel."
                )
                excel_ingester = ExcelConverter(
                    config=config,
                    category_info=category_info,
                )
                excel_ingester(objects)

            else:
                logger.info(
                    "Attempting to process metadata and convert to a single dataframe."
                )
                ingester = SinglePandasTableConverter(
                    config=config,
                    category_info=category_info,
                )
                ingester(objects)

    @classmethod
    def get_items(
        cls,
        api: ElabftwApi,
        ids: Optional[Collection[int]] = None,
        category_ids: Optional[Iterable[int]] = None,
    ) -> Iterator[Item]:
        """Iterates over categories and yields iterators over the items in those categories"""
        if category_ids is None:
            return api.items.iter(ids=ids)
        else:
            return chain(
                *(api.items.iter(ids=ids, cat=cat_id) for cat_id in category_ids)
            )

    @classmethod
    def get_experiments(
        cls,
        api: ElabftwApi,
        ids: Optional[Collection[int]] = None,
        template_names: Optional[Sequence[str]] = None,
    ) -> Iterator[Experiment]:
        experiments = api.experiments.iter(ids=ids)
        if not template_names:
            for experiment in experiments:
                yield experiment
        else:
            for experiment in experiments:
                for template_name in template_names:
                    if template_name in experiment.title:
                        yield experiment
                        break


class JSONConverter(NamedTuple):
    """Converts API objects to JSON and saves them to a file"""

    config: IngestConfiguration

    def __call__(
        self,
        objects: Iterable[Dictable],
    ) -> None:
        transformed = self._transform(objects)
        self._save(transformed)

    def _transform(
        self,
        objects: Iterable[Dictable],
    ) -> Iterator[str]:
        indent = 2 if self.config.indent else None
        transformer = JSONTransformer.new(indent=indent)
        lines = transformer(objects)
        return lines

    def _save(
        self,
        lines: Iterable[str],
    ) -> None:
        outputter = LineWriter.new(self.config.output)
        outputter(lines=lines)


class SinglePandasTableConverter(NamedTuple):
    """Convert API objects to a pandas DF and save to a file"""

    config: IngestConfiguration
    category_info: CategoryInfo

    def __call__(
        self,
        objects: Iterable[Dictable],
    ) -> None:
        transformed = self._transform(objects)
        self._save(transformed)

    def _transform(
        self,
        objects: Iterable[Dictable],
    ) -> pd.DataFrame:
        raw_table = PandasDataFrameTransformer.new()(objects)
        if not self.config.expand_metadata:
            return raw_table

        if self.config.table_shape == TableShapes.WIDE:
            logger.info("Deriving metadata schema for wide format")
            metadata_schema = self._get_category_metadata_schema()
            transformer = PandasDataFrameMetadataTransformer.for_wide_table(
                object_type=self.config.object_type,
                cell_content=self.config.cell_content,
                metadata_schema=metadata_schema,
                sanitize_column_names=self.config.sanitize_column_names,
            )
        else:
            transformer = PandasDataFrameMetadataTransformer.for_long_table(
                object_type=self.config.object_type,
            )
        return transformer(raw_table)

    def _save(self, df: pd.DataFrame) -> None:
        single_writer = PandasWriter.new(
            self.config.output,
            format=self.config.format,
        )
        single_writer(df)

    def _get_category_metadata_schema(self) -> Optional[dict[str, str]]:
        if self.config.use_template_metadata_schema:
            category_ids = self.category_info.get_ids(self.config.categories)
            logger.debug("We need category metadata based on ids: %s" % category_ids)
            if len(category_ids) == 1:
                schema = self.category_info.get_metadata_schema(
                    cell_content=self.config.cell_content,
                    category_id=category_ids[0],
                )
                logger.debug("Metadata schema:\n%s" % json.dumps(schema, indent=2))
                return schema
            else:
                raise ValueError(
                    "Can not derive a single metadata schema with no or multiple template ids provided."
                )
        else:
            logger.debug("No metadata schema required since we want all columns.")
            return None


class ExcelConverter(NamedTuple):
    """Converts API objects to multiple dataframes and saves them to an Excel"""

    config: IngestConfiguration
    category_info: CategoryInfo

    def __call__(
        self,
        objects: Iterable[Dictable],
    ) -> None:
        split_tables = self._transform(objects)
        self._save(split_tables)

    def _transform(
        self,
        objects: Iterable[Dictable],
    ) -> Iterator[SplitDataFrame]:
        raw_table = PandasDataFrameTransformer.new()(objects)
        if not self.config.expand_metadata:
            multi_tables = MultiPandasDataFrameTransformer.for_raw_tables(
                object_type=self.config.object_type,
                categories=self.config.categories,
            )
        else:
            if self.config.table_shape == TableShapes.WIDE:
                categories_metadata_schema = self._get_categories_metadata_schemas()
                multi_tables = MultiPandasDataFrameTransformer.for_wide_tables(
                    object_type=self.config.object_type,
                    categories_metadata_schema=categories_metadata_schema,
                    cell_content=self.config.cell_content,
                    sanitize_column_names=self.config.sanitize_column_names,
                )
            else:
                multi_tables = MultiPandasDataFrameTransformer.for_long_tables(
                    object_type=self.config.object_type,
                    categories=self.config.categories,
                )
        return multi_tables(raw_table)

    def _save(
        self,
        dataframes: Iterable[SplitDataFrame],
    ) -> None:
        writer = ExcelWriter.new(self.config.output)
        writer(dataframes)

    def _get_categories_metadata_schemas(
        self,
    ) -> Optional[dict[str, Optional[dict[str, str]]]]:
        if self.config.use_template_metadata_schema:
            logger.debug("We only want template metadata columns, not all.")
            category_ids = self.category_info.get_ids(self.config.categories)
            return self.category_info.get_metadata_schemas_by_name(
                cell_content=self.config.cell_content,
                category_ids=category_ids,
            )
        else:
            logger.debug("No metadata schema required since we want all metadata.")
            if self.config.categories:
                return {category: None for category in self.config.categories}
            else:
                return None


class CategoryInfo:
    """Retrieves and stores information about categories/templates"""

    def __init__(
        self,
        getter: Callable[[], Iterable[Category]],
    ) -> None:
        self.getter = getter

    @cached_property
    def categories_map(self) -> dict[int, Category]:
        return self.fetch_categories_map(self.getter)

    @cached_property
    def ids(self) -> list[int]:
        return list(self.categories_map.keys())

    def get_ids(self, categories: Optional[Iterable[str]]) -> list[int]:
        if not categories:
            return self.ids
        else:
            ids = []
            for category in categories:
                if category not in self.name_to_ids:
                    raise KeyError(
                        f"Category '{category}' does not exist. "
                        f"Valid options: {list(self.name_to_ids.keys())}"
                    )
                ids.extend(self.name_to_ids[category])
            return ids

    @cached_property
    def name_to_ids(self) -> dict[str, list[int]]:
        name_to_ids: dict[str, list[int]] = {}
        for category_id, category in self.categories_map.items():
            name = category.title
            if name not in name_to_ids:
                name_to_ids[name] = list()
            else:
                logger.warning(
                    "The categegory name '%s' appears in multiple category ids" % name
                )
            name_to_ids[name].append(category_id)

        return name_to_ids

    @cached_property
    def id_to_name(self) -> dict[int, str]:
        return {
            category_id: category.title
            for category_id, category in self.categories_map.items()
        }

    def get_metadata_schemas_by_name(
        self,
        cell_content: TableCellContentType,
        category_ids: Optional[Collection[int]] = None,
    ) -> dict[str, Optional[dict[str, str]]]:
        """Get the metadata schemas from multiple categories."""
        if category_ids is None:
            category_ids = self.categories_map.keys()

        categories_metadata_schema: dict[str, Optional[dict[str, Any]]] = {}

        for category_id in category_ids:
            category = self.categories_map[category_id]
            if category.title in categories_metadata_schema:
                raise RuntimeError(
                    f"The category name '{category.title}' is duplicated, "
                    "can not create a mapping based on category names."
                )

            schema = self.get_metadata_schema(
                cell_content=cell_content,
                category_id=category_id,
            )
            categories_metadata_schema[category.title] = schema
        return categories_metadata_schema

    def get_metadata_schemas(
        self,
        cell_content: TableCellContentType,
        category_ids: Optional[Collection[int]] = None,
    ) -> dict[int, Optional[dict[str, str]]]:
        """Get the metadata schemas from multiple categories."""
        if category_ids is None:
            category_ids = self.categories_map.keys()

        categories_metadata_schema: dict[int, Optional[dict[str, Any]]] = {}

        for category_id in category_ids:
            schema = self.get_metadata_schema(
                cell_content=cell_content,
                category_id=category_id,
            )
            categories_metadata_schema[category_id] = schema
        return categories_metadata_schema

    def get_metadata_schema(
        self,
        cell_content: TableCellContentType,
        category_id: int,
    ) -> Optional[dict[str, Any]]:
        """Get the metadata schema from a known category id"""
        if category_id not in self.categories_map:
            raise KeyError(f"Category id {category_id} not found.")
        category = self.categories_map[category_id]
        schema = self.parse_schema_from_metadata(
            metadata=category.metadata,
            cell_content=cell_content,
        )
        return schema

    @classmethod
    def parse_schema_from_metadata(
        cls,
        metadata: Optional[str],
        cell_content: TableCellContentType,
    ) -> dict[str, str]:
        """Get the schema of the object metadata in wide table format from the
        category metadata string"""
        parsed_metadata = MetadataParser()(metadata)
        transformer = ParsedMetadataToPandasDtype.new(cell_content)
        return transformer(parsed_metadata)

    @classmethod
    def fetch_categories_map(
        cls,
        getter: Callable[[], Iterable[Category]],
    ) -> dict[int, Category]:
        """Get a mapping of category ids to category data"""
        logger.info("Fetching categories information...")
        category_map = {category.id: category for category in getter()}
        logger.info("Returned %s categories" % len(category_map))
        logger.debug("Category ids: %s" % list(category_map.keys()))
        logger.debug(
            "Category titles: %s"
            % [category.title for category in category_map.values()]
        )
        return category_map



from enum import Enum
from pathlib import Path

import git
from pydantic import BaseModel

from elabftwcontrol._logging import logger
from elabftwcontrol.client import ElabftwApi
from elabftwcontrol.upload.jobs import WorkEvaluator
from elabftwcontrol.upload.parsers import ManifestParser
from elabftwcontrol.upload.state import State


class ApplyJobType(str, Enum):
    CREATE = "create"
    DESTROY = "destroy"


class ApplyJobConfig(BaseModel):
    jobtype: ApplyJobType
    manifest_folder: Path | str
    ignore_state: bool
    state_file: Path | str | None
    version: str | None
    n_retries: int
    no_prompt: bool
    dry_run: bool


class ApplyJob:
    def __init__(
        self,
        api: ElabftwApi,
        config: ApplyJobConfig,
    ) -> None:
        self.api = api
        self.config = config

    def __call__(self) -> None:
        parser = ManifestParser(self.config.manifest_folder)
        manifest_index = parser.parse()

        state: State
        if self.config.ignore_state:
            logger.debug("Ignoring state.")
            state = State.empty()
        else:
            if self.config.state_file is None:
                logger.debug("Pulling state from the API...")
                state = State.from_api(self.api, skip_untracked=True)
            else:
                logger.debug("Getting state from a file.")
                state = State.from_file(self.config.state_file)

        version: str = ""
        if self.config.jobtype == ApplyJobType.CREATE:
            if self.config.version is None:
                try:
                    repo = git.Repo(
                        self.config.manifest_folder,
                        search_parent_directories=True,
                    )
                    version = repo.head.object.hexsha
                    assert version is not None
                    logger.info(
                        "Version %s was detected from git and will be used in the apply."
                        % version
                    )
                except Exception:
                    logger.warning(
                        "No version was passed or could be detected from version control."
                    )
                    version = ""
            else:
                version = self.config.version
                logger.info("Using version %s in the apply." % version)

        evaluator = WorkEvaluator.new(
            manifest_index=manifest_index,
            elab_state=state,
            version=version,
        )

        if self.config.jobtype == ApplyJobType.CREATE:
            plan = evaluator.evaluate_apply()
        elif self.config.jobtype == ApplyJobType.DESTROY:
            plan = evaluator.evaluate_destroy()
        else:
            raise RuntimeError(f"Could not identify job type: {self.config.jobtype}")

        if not plan:
            print("No work to be done!")
            return

        print("The following is the work to be done:")
        print()

        plan.print_to_console()

        print()
        print()

        if not self.config.no_prompt:
            if input("Do you wish to apply the changes? [Y|n] ").startswith("n"):
                return

        if not self.config.dry_run:
            plan.execute(
                client=self.api,
                start=0,
                n_retries=self.config.n_retries,
            )

        print("Completed apply.")



from __future__ import annotations

from enum import Enum


class ElabObjectType(str, Enum):
    ITEM = "item"
    ITEMS_TYPE = "items_type"
    EXPERIMENT = "experiment"
    EXPERIMENTS_TEMPLATE = "experiments_template"


class EntityTypes(str, Enum):
    ITEM = "items"
    ITEMS_TYPE = "items_types"
    EXPERIMENT = "experiments"
    EXPERIMENTS_TEMPLATE = "experiments_templates"


class SingleObjectTypes(str, Enum):
    ITEM = EntityTypes.ITEM.value
    EXPERIMENT = EntityTypes.EXPERIMENT.value
